# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper classes and functions for evaluation tests."""

from langfun.core import language_model
from langfun.core import llms
from langfun.core import message as message_lib
from langfun.core import structured

from langfun.core.eval.v2 import evaluation as evaluation_lib
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
from langfun.core.eval.v2 import metrics as metrics_lib

import pyglove as pg

Example = example_lib.Example
Suite = experiment_lib.Suite
Evaluation = evaluation_lib.Evaluation
RunId = experiment_lib.RunId
Run = experiment_lib.Run


@pg.functor()
def test_inputs(num_examples: int | None = 10):
  if num_examples is None:
    num_examples = 20
  return [
      pg.Dict(x=i, y=i ** 2, groundtruth=i + i ** 2)
      for i in range(num_examples)
  ]


class TestLLM(llms.Fake):
  """Test language model."""

  offset: int = 0

  def _response_from(self, prompt: message_lib.Message) -> message_lib.Message:
    return message_lib.AIMessage(
        str(prompt.metadata.x + prompt.metadata.y + self.offset)
    )

  @property
  def resource_id(self) -> str:
    return f'test_llm:{self.offset}'


class TestEvaluation(Evaluation):
  """Test evaluation class."""
  inputs = test_inputs()
  metrics = [metrics_lib.Match()]
  lm: language_model.LanguageModel = TestLLM()

  def process(self, example):
    v = example.input
    if v.x == 5:
      raise ValueError('x should not be 5')
    return structured.query(
        '{{x}} + {{y}} = ?', int, lm=self.lm, x=v.x, y=v.y,
        metadata_x=v.x, metadata_y=v.y
    )


class BadJsonConvertible(pg.Object):

  def to_json(self, *args, **kwargs):
    raise ValueError('Cannot convert to JSON.')


class TestEvaluationWithExampleCheckpointingError(TestEvaluation):
  """Test evaluation class with bad example checkpointing."""
  inputs = test_inputs()
  metrics = [metrics_lib.Match()]

  def process(self, example):
    return 1, dict(
        x=BadJsonConvertible()
    )


class BadHtmlConvertible(pg.Object, pg.views.HtmlTreeView.Extension):

  def _html_tree_view(self, *args, **kwargs):
    raise ValueError('Cannot render HTML.')


class TestEvaluationWithExampleHtmlGenerationError(Evaluation):
  """Test evaluation class with bad example HTML generation."""
  inputs = test_inputs()
  metrics = [metrics_lib.Match()]

  def process(self, example):
    return 1, dict(
        x=BadHtmlConvertible()
    )


class TestEvaluationWithIndexHtmlGenerationError(TestEvaluation):
  """Test evaluation class with bad index HTML generation."""

  def _html_tree_view(self, *args, **kwargs):
    raise ValueError('Cannot render HTML.')


def test_experiment():
  """Returns a test experiment."""
  return Suite([
      TestEvaluation(lm=TestLLM(offset=0)),
      TestEvaluation(lm=TestLLM(offset=pg.oneof(range(5)))),
  ])


def test_experiment_with_example_checkpointing_error():
  """Returns a test experiment with example checkpointing error."""
  return TestEvaluationWithExampleCheckpointingError()


def test_experiment_with_example_html_generation_error():
  """Returns a test experiment with bad example HTML."""
  return TestEvaluationWithExampleHtmlGenerationError()


def test_experiment_with_index_html_generation_error():
  """Returns a test experiment with bad index HTML."""
  return TestEvaluationWithIndexHtmlGenerationError()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Natural language utilities."""

import abc
import pyglove as pg


class NaturalLanguageFormattable(pg.Formattable):
  """Base class for natural language formattable objects."""

  @abc.abstractmethod
  def natural_language_format(self) -> str:
    """Returns the natural language format of current object."""

  def format(
      self,
      *args,
      natural_language: bool = False,
      **kwargs
  ) -> str:
    if natural_language:
      return self.natural_language_format()
    return super().format(*args, **kwargs)

  def __str__(self):
    return self.natural_language_format()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""In-memory LM cache."""

import collections
import contextlib
import json
from typing import Annotated, Any, Iterator
import langfun.core as lf
from langfun.core.llms.cache import base
import pyglove as pg


@pg.use_init_args(['filename', 'ttl', 'key'])
class InMemory(base.LMCacheBase):
  """In memory cache."""

  filename: Annotated[
      str | None,
      (
          'File name to load and save in memory cache.'
      )
  ] = None

  def _on_bound(self) -> None:
    super()._on_bound()
    self._cache = collections.defaultdict(dict)

    if self.filename is not None:
      try:
        records = pg.load(self.filename)
        for record in records:
          model_cache = {}
          for entry in record.entries:
            model_cache[entry.k] = entry.v
          self._cache[record.model_id] = model_cache
      except FileNotFoundError:
        pg.logging.warning(
            "Creating a new cache as cache file '%s' does not exist.",
            self.filename,
        )
      except json.JSONDecodeError:
        pg.logging.warning(
            "Creating a new cache as cache file '%s' is corrupted.",
            self.filename,
        )

  def model_ids(self) -> list[str]:
    """Returns the model ids of cached queires."""
    return list(self._cache.keys())

  def __len__(self) -> int:
    """Returns the number of entries in the cache."""
    return sum(len(v) for v in self._cache.values())

  def keys(self, model_id: str | None = None) -> Iterator[str]:
    """Returns the cached keys for a model."""
    if model_id is None:
      for model_cache in self._cache.values():
        for k in model_cache.keys():
          yield k
    else:
      for k in self._cache[model_id].keys():
        yield k

  def values(self, model_id: str | None = None) -> Iterator[base.LMCacheEntry]:
    """Returns the cached entries for a model."""
    if model_id is None:
      for model_cache in self._cache.values():
        for v in model_cache.values():
          yield v
    else:
      for v in self._cache[model_id].values():
        yield v

  def items(
      self,
      model_id: str | None = None
      ) -> Iterator[tuple[str, base.LMCacheEntry]]:
    """Returns the cached items for a model."""
    if model_id is None:
      for model_cache in self._cache.values():
        for k, v in model_cache.items():
          yield k, v
    else:
      for k, v in self._cache[model_id].items():
        yield k, v

  def _get(self, model_id: str, key: Any) -> base.LMCacheEntry | None:
    """Returns a LM cache entry associated with the key."""
    return self._cache[model_id].get(key, None)

  def _put(self, model_id: str, key: Any, entry: base.LMCacheEntry) -> None:
    """Puts a LM cache entry associated with the key."""
    self._cache[model_id][key] = entry

  def _delete(self, model_id: str, key: str) -> bool:
    """Deletes a LM cache entry associated with the key."""
    model_cache = self._cache.get(model_id, None)
    if model_cache is None:
      return False
    return model_cache.pop(key, None) is not None

  def reset(self, model_id: str | None = None) -> None:
    """Resets the cache."""
    if model_id is not None:
      self._cache[model_id].clear()
    else:
      self._cache.clear()

  def _sym_clone(self, deep: bool, memo: Any = None) -> 'InMemory':
    v = super()._sym_clone(deep, memo)
    v._cache = self._cache  # pylint: disable=protected-access
    return v

  def save(self, path: str | None = None) -> None:
    """Saves the in-memory cache."""
    if path is None:
      if self.filename is None:
        raise ValueError('`path` must be specified.')
      path = self.filename

    # Do nothing if there is no update, this avoids unnecessary rewrites.
    if self.stats.num_updates == 0 and path == self.filename:
      return

    records = []
    for model_id in self.model_ids():
      entries = [dict(k=k, v=v) for k, v in self.items(model_id)]
      records.append(dict(model_id=model_id, entries=entries))
    pg.save(records, path)


@contextlib.contextmanager
def lm_cache(filename: str | None = None) -> Iterator[InMemory]:
  """Context manager to enable cache for LMs under the context.

  If LMs under the context manager have explicitly specified cache, they will
  use their own cache. Otherwise they will use the cache created by the context
  manager.

  Args:
    filename: If not None, JSON file to load and save the cache.

  Yields:
    A cache object created.
  """
  cache = InMemory(filename)
  try:
    with lf.context(cache=cache):
      yield cache
  finally:
    if filename is not None:
      cache.save()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun LLM implementations."""

# pylint: disable=g-importing-member
# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top

# LMs for testing.
from langfun.core.llms.fake import Fake
from langfun.core.llms.fake import Echo
from langfun.core.llms.fake import Pseudo
from langfun.core.llms.fake import StaticMapping
from langfun.core.llms.fake import StaticResponse
from langfun.core.llms.fake import StaticSequence

# Compositional models.
from langfun.core.llms.compositional import RandomChoice

# Base models by request/response protocol.
from langfun.core.llms.rest import REST
from langfun.core.llms.openai_compatible import OpenAICompatible
from langfun.core.llms.gemini import Gemini
from langfun.core.llms.anthropic import Anthropic

# Base models by serving platforms.
from langfun.core.llms.vertexai import VertexAI
from langfun.core.llms.groq import Groq
from langfun.core.llms.azure_openai import AzureOpenAI

# Gemini models.
from langfun.core.llms.google_genai import GenAI
from langfun.core.llms.google_genai import Gemini25Pro
from langfun.core.llms.google_genai import Gemini25Flash
from langfun.core.llms.google_genai import Gemini25ProPreview_20250605
from langfun.core.llms.google_genai import Gemini25FlashPreview_20250520
from langfun.core.llms.google_genai import Gemini25ProPreview_20250506
from langfun.core.llms.google_genai import Gemini25FlashPreview_20250417
from langfun.core.llms.google_genai import Gemini25ProPreview_20250325
from langfun.core.llms.google_genai import Gemini2Flash
from langfun.core.llms.google_genai import Gemini2Flash_001
from langfun.core.llms.google_genai import Gemini2FlashLitePreview_20250205
from langfun.core.llms.google_genai import Gemini15Pro
from langfun.core.llms.google_genai import Gemini15Pro_002
from langfun.core.llms.google_genai import Gemini15Pro_001
from langfun.core.llms.google_genai import Gemini15Flash
from langfun.core.llms.google_genai import Gemini15Flash_002
from langfun.core.llms.google_genai import Gemini15Flash_001
from langfun.core.llms.google_genai import Gemini15Flash8B
from langfun.core.llms.google_genai import Gemini15Flash8B_001
from langfun.core.llms.google_genai import Gemini2ProExp_20250205
from langfun.core.llms.google_genai import Gemini2FlashThinkingExp_20250121
from langfun.core.llms.google_genai import GeminiExp_20241206
from langfun.core.llms.google_genai import Gemini25FlashImagePreview

from langfun.core.llms.vertexai import VertexAIGemini
from langfun.core.llms.vertexai import VertexAIGemini2Flash
from langfun.core.llms.vertexai import VertexAIGemini2Flash_001
from langfun.core.llms.vertexai import VertexAIGemini2FlashLitePreview_20250205
from langfun.core.llms.vertexai import VertexAIGemini15Pro
from langfun.core.llms.vertexai import VertexAIGemini15Pro_002
from langfun.core.llms.vertexai import VertexAIGemini15Pro_001
from langfun.core.llms.vertexai import VertexAIGemini15Flash
from langfun.core.llms.vertexai import VertexAIGemini15Flash_002
from langfun.core.llms.vertexai import VertexAIGemini15Flash_001
from langfun.core.llms.vertexai import VertexAIGemini15Flash8B
from langfun.core.llms.vertexai import VertexAIGemini15Flash8B_001

from langfun.core.llms.vertexai import VertexAIGemini2ProExp_20250205
from langfun.core.llms.vertexai import VertexAIGemini2FlashThinkingExp_20250121
from langfun.core.llms.vertexai import VertexAIGeminiExp_20241206
from langfun.core.llms.vertexai import VertexAIGemini25ProExp_20250325
from langfun.core.llms.vertexai import VertexAIGemini25ProPreview_20250325
from langfun.core.llms.vertexai import VertexAIGemini25FlashPreview_20250417
from langfun.core.llms.vertexai import VertexAIGemini25ProPreview_20250506
from langfun.core.llms.vertexai import VertexAIGemini25FlashPreview_20250520
from langfun.core.llms.vertexai import VertexAIGemini25ProPreview_20250605
from langfun.core.llms.vertexai import VertexAIGemini25Pro
from langfun.core.llms.vertexai import VertexAIGemini25Flash
from langfun.core.llms.vertexai import VertexAIGemini25FlashImagePreview

# For backward compatibility.
GeminiPro1_5 = Gemini15Pro
GeminiFlash1_5 = Gemini15Flash
VertexAIGeminiPro1_5 = VertexAIGemini15Pro
VertexAIGeminiFlash1_5 = VertexAIGemini15Flash

# OpenAI models.
from langfun.core.llms.openai import OpenAI

from langfun.core.llms.openai import Gpt41
from langfun.core.llms.openai import GptO3
from langfun.core.llms.openai import GptO4Mini
from langfun.core.llms.openai import Gpt45Preview_20250227
from langfun.core.llms.openai import GptO3Mini
from langfun.core.llms.openai import GptO1
from langfun.core.llms.openai import GptO1Preview
from langfun.core.llms.openai import GptO1Preview_20240912
from langfun.core.llms.openai import GptO1Mini
from langfun.core.llms.openai import GptO1Mini_20240912

from langfun.core.llms.openai import Gpt4oMini
from langfun.core.llms.openai import Gpt4oMini_20240718
from langfun.core.llms.openai import Gpt4o
from langfun.core.llms.openai import Gpt4o_20241120
from langfun.core.llms.openai import Gpt4o_20240806
from langfun.core.llms.openai import Gpt4o_20240513

from langfun.core.llms.openai import Gpt4Turbo
from langfun.core.llms.openai import Gpt4Turbo_20240409
from langfun.core.llms.openai import Gpt4TurboPreview
from langfun.core.llms.openai import Gpt4TurboPreview_20240125
from langfun.core.llms.openai import Gpt4TurboPreview_20231106
from langfun.core.llms.openai import Gpt4
from langfun.core.llms.openai import Gpt4_20230613

from langfun.core.llms.openai import Gpt4_32K
from langfun.core.llms.openai import Gpt4_32K_20230613

from langfun.core.llms.openai import Gpt35Turbo
from langfun.core.llms.openai import Gpt35Turbo_20240125
from langfun.core.llms.openai import Gpt35Turbo_20231106
from langfun.core.llms.openai import Gpt35Turbo_20230613
from langfun.core.llms.openai import Gpt35Turbo16K
from langfun.core.llms.openai import Gpt35Turbo16K_20230613

# For backward compatibility.
Gpt4TurboPreview_0125 = Gpt4TurboPreview_20240125
Gpt4TurboPreview_1106 = Gpt4TurboPreview_20231106
Gpt4_0613 = Gpt4_20230613
Gpt4_32K_0613 = Gpt4_32K_20230613
Gpt35Turbo_0125 = Gpt35Turbo_20240125
Gpt35Turbo_1106 = Gpt35Turbo_20231106
Gpt35Turbo_0613 = Gpt35Turbo_20230613
Gpt35Turbo16K_0613 = Gpt35Turbo16K_20230613

from langfun.core.llms.openai import Gpt35

# Anthropic models.

from langfun.core.llms.anthropic import Claude4
from langfun.core.llms.anthropic import Claude4Sonnet_20250514
from langfun.core.llms.anthropic import Claude4Opus_20250514
from langfun.core.llms.anthropic import Claude37
from langfun.core.llms.anthropic import Claude37Sonnet_20250219
from langfun.core.llms.anthropic import Claude35Sonnet
from langfun.core.llms.anthropic import Claude35Sonnet_20241022
from langfun.core.llms.anthropic import Claude35Haiku
from langfun.core.llms.anthropic import Claude35Haiku_20241022
from langfun.core.llms.anthropic import Claude3Opus
from langfun.core.llms.anthropic import Claude3Opus_20240229
from langfun.core.llms.anthropic import Claude3Sonnet
from langfun.core.llms.anthropic import Claude3Sonnet_20240229
from langfun.core.llms.anthropic import Claude3Haiku
from langfun.core.llms.anthropic import Claude3Haiku_20240307

from langfun.core.llms.vertexai import VertexAIAnthropic
from langfun.core.llms.vertexai import VertexAIClaude4Opus_20250514
from langfun.core.llms.vertexai import VertexAIClaude4Sonnet_20250514
from langfun.core.llms.vertexai import VertexAIClaude37Sonnet_20250219
from langfun.core.llms.vertexai import VertexAIClaude35Sonnet_20241022
from langfun.core.llms.vertexai import VertexAIClaude35Haiku_20241022
from langfun.core.llms.vertexai import VertexAIClaude3Opus_20240229

# Misc open source models.

# Gemma models.
from langfun.core.llms.groq import GroqGemma2_9B_IT

# Llama models.
from langfun.core.llms.vertexai import VertexAILlama
from langfun.core.llms.vertexai import VertexAILlama32_90B
from langfun.core.llms.vertexai import VertexAILlama31_405B
from langfun.core.llms.vertexai import VertexAILlama31_70B
from langfun.core.llms.vertexai import VertexAILlama31_8B

from langfun.core.llms.groq import GroqLlama33_70B_Versatile
from langfun.core.llms.groq import GroqLlama33_70B_SpecDec
from langfun.core.llms.groq import GroqLlama32_1B
from langfun.core.llms.groq import GroqLlama32_3B
from langfun.core.llms.groq import GroqLlama32_11B_Vision
from langfun.core.llms.groq import GroqLlama32_90B_Vision

# Mistral models.
from langfun.core.llms.vertexai import VertexAIMistral
from langfun.core.llms.vertexai import VertexAIMistralLarge_20241121
from langfun.core.llms.vertexai import VertexAICodestral_20250113

from langfun.core.llms.groq import GroqMistral_8x7B

# DeepSeek models.
from langfun.core.llms.deepseek import DeepSeek
from langfun.core.llms.deepseek import DeepSeekV3
from langfun.core.llms.deepseek import DeepSeekR1

# LLaMA C++ models.
from langfun.core.llms.llama_cpp import LlamaCppRemote

# Placeholder for Google-internal imports.

# Include cache as sub-module.
from langfun.core.llms import cache

# pylint: enable=g-import-not-at-top
# pylint: enable=g-bad-import-order
# pylint: enable=g-importing-member



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compositional models."""
import unittest

import langfun.core as lf
from langfun.core.llms import compositional
from langfun.core.llms import fake


class RandomChoiceTest(unittest.TestCase):

  def test_basic(self):
    lm = compositional.RandomChoice([
        fake.StaticResponse('hi'),
        fake.StaticSequence(['hello', 'world'])
    ])
    self.assertEqual(
        lm.model_id,
        'RandomChoice(StaticResponse, StaticSequence)'
    )
    self.assertEqual(
        lm.resource_id,
        'RandomChoice(StaticResponse, StaticSequence)'
    )
    self.assertEqual(
        [lm('a'), lm('b'), lm('c')],
        ['hello', 'world', 'hi']
    )
    lm = lm.clone()
    self.assertEqual(
        [
            x.samples[0].response for x in [
                lm.sample(['a'])[0],
                lm.sample(['b'])[0],
                lm.sample(['c'])[0],
            ]
        ],
        ['hello', 'world', 'hi']
    )
    self.assertEqual(
        lm.score('hello', ['world']),
        [lf.LMScoringResult(0.0)]
    )
    self.assertEqual(
        lm.tokenize('hello'),
        [('hello', 0)]
    )

  def test_sampling_options(self):
    lm = compositional.RandomChoice([
        fake.StaticResponse('hi'),
        fake.StaticSequence(['hello', 'world'])
    ], temperature=0.5)
    self.assertEqual(
        lm.candidates[0].sampling_options.temperature,
        0.5
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for base action."""

import asyncio
import time
import unittest

import langfun.core as lf
from langfun.core.agentic import action as action_lib
from langfun.core.llms import fake
import langfun.core.structured as lf_structured
import pyglove as pg


class Bar(action_lib.Action):
  simulate_action_error: bool = False
  simulate_execution_time: float = 0

  def call(self, session, *, lm, **kwargs):
    assert session.current_action.action is self
    session.info('Begin Bar')
    time.sleep(self.simulate_execution_time)
    session.query('bar', lm=lm)
    session.add_metadata(note='bar')
    if self.simulate_action_error:
      raise ValueError('Bar error')
    return 2 + pg.contextual_value('baz', 0)


class Foo(action_lib.Action):
  x: int
  simulate_action_error: bool = False
  simulate_query_error: bool = False
  simulate_execution_time: list[float] = [0, 0, 0, 0]
  max_bar_execution_time: float | None = None

  def call(self, session, *, lm, **kwargs):
    assert session.current_action.action is self
    with session.track_phase('prepare'):
      session.info('Begin Foo', x=1)
      time.sleep(self.simulate_execution_time[0])
      session.query(
          'foo',
          schema=int if self.simulate_query_error else None,
          lm=lm
      )
    with session.track_queries():
      time.sleep(self.simulate_execution_time[1])
      self.make_additional_query(lm)
    session.add_metadata(note='foo')

    def _sub_task(i):
      session.add_metadata(**{f'subtask_{i}': i})
      time.sleep(self.simulate_execution_time[2])
      return lf_structured.query(f'subtask_{i}', lm=lm)

    for i, output, error in session.concurrent_map(
        _sub_task, range(3), max_workers=2, silence_on_errors=None,
    ):
      assert isinstance(i, int), i
      assert isinstance(output, str), output
      assert error is None, error
    return self.x + Bar(
        simulate_action_error=self.simulate_action_error,
        simulate_execution_time=self.simulate_execution_time[3]
    )(session, lm=lm, max_execution_time=self.max_bar_execution_time)

  def make_additional_query(self, lm):
    lf_structured.query('additional query', lm=lm)


class ActionInvocationTest(unittest.TestCase):

  def test_basics(self):
    action_invocation = action_lib.ActionInvocation(
        action=Foo(1)
    )
    self.assertEqual(action_invocation.id, '')
    root = action_lib.ActionInvocation(action=action_lib.RootAction())
    root.execution.append(action_invocation)
    self.assertIs(action_invocation.parent_action, root)
    self.assertEqual(action_invocation.id, '/a1')


class ExecutionTraceTest(unittest.TestCase):

  def test_basics(self):
    execution = action_lib.ExecutionTrace()
    self.assertEqual(execution.id, '')

    root = action_lib.ActionInvocation(action=action_lib.RootAction())
    action_invocation = action_lib.ActionInvocation(
        action=Foo(1)
    )
    root.execution.append(action_invocation)
    self.assertEqual(action_invocation.execution.id, '/a1')

    root.execution.reset()
    self.assertFalse(root.execution)


class SessionTest(unittest.TestCase):

  def test_succeeded_trajectory(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1)
    self.assertIsNone(foo.session)
    self.assertIsNone(foo.invocation)
    self.assertIsNone(foo.result)
    self.assertIsNone(foo.metadata)

    session = action_lib.Session(id='agent@1')
    self.assertEqual(session.id, 'agent@1')
    self.assertFalse(session.has_started)
    self.assertFalse(session.has_stopped)

    # Render HTML view to trigger dynamic update during execution.
    _ = session.to_html()

    with session:
      result = foo(session, lm=lm, verbose=True)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(result, 3)
    self.assertIsNone(foo.session)
    self.assertEqual(foo.result, 3)
    self.assertEqual(
        foo.metadata, dict(note='foo', subtask_0=0, subtask_1=1, subtask_2=2)
    )

    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)

    #
    # Inspecting the root invocation.
    #

    root = session.root
    self.assertIsNone(root.parent_action)
    self.assertEqual(root.id, 'agent@1:')
    self.assertEqual(root.execution.id, 'agent@1:')
    self.assertEqual(len(root.execution), 1)
    self.assertIs(root.execution[0].action, foo)

    self.assertTrue(root.execution.has_started)
    self.assertTrue(root.execution.has_stopped)
    self.assertGreater(root.execution.elapse, 0)
    self.assertEqual(root.result, 3)
    self.assertEqual(
        root.metadata,
        dict(note='foo', subtask_0=0, subtask_1=1, subtask_2=2)
    )

    # The root space should have one action (foo), no queries, and no logs.
    self.assertEqual(len(root.actions), 1)
    self.assertEqual(len(root.queries), 0)
    self.assertEqual(len(root.logs), 0)
    # 1 query from Bar, 2 from Foo and 3 from parallel executions.
    self.assertEqual(len(session.all_queries), 6)
    self.assertEqual(len(root.all_queries), 6)
    # 2 actions: Foo and Bar.
    self.assertEqual(len(session.all_actions), 2)
    self.assertEqual(len(root.all_actions), 2)
    # 1 log from Bar and 1 from Foo.
    self.assertEqual(len(session.all_logs), 2)
    self.assertEqual(len(root.all_logs), 2)
    self.assertIs(session.usage_summary, root.usage_summary)
    self.assertEqual(root.usage_summary.total.num_requests, 6)

    # Inspecting the top-level action (Foo)
    foo_invocation = root.execution[0]
    self.assertIs(foo.invocation, foo_invocation)
    self.assertIs(foo_invocation.parent_action, root)
    self.assertEqual(foo_invocation.id, 'agent@1:/a1')
    self.assertEqual(foo_invocation.execution.id, 'agent@1:/a1')
    self.assertEqual(len(foo_invocation.execution.items), 4)

    # Prepare phase.
    prepare_phase = foo_invocation.execution[0]
    self.assertIsInstance(prepare_phase, action_lib.ExecutionTrace)
    self.assertEqual(prepare_phase.id, 'agent@1:/a1/prepare')
    self.assertEqual(len(prepare_phase.items), 2)
    self.assertTrue(prepare_phase.has_started)
    self.assertTrue(prepare_phase.has_stopped)
    self.assertEqual(prepare_phase.usage_summary.total.num_requests, 1)
    self.assertIsInstance(prepare_phase.items[0], lf.logging.LogEntry)
    self.assertIsInstance(prepare_phase.items[1], lf_structured.QueryInvocation)
    self.assertEqual(prepare_phase.items[1].id, 'agent@1:/a1/prepare/q1')

    # Tracked queries.
    query_invocation = foo_invocation.execution[1]
    self.assertIsInstance(query_invocation, lf_structured.QueryInvocation)
    self.assertEqual(query_invocation.id, 'agent@1:/a1/q2')
    self.assertIs(query_invocation.lm, lm)
    self.assertEqual(
        foo_invocation.execution.indexof(
            query_invocation, lf_structured.QueryInvocation
        ),
        1
    )
    self.assertEqual(
        root.execution.indexof(
            query_invocation, lf_structured.QueryInvocation
        ),
        -1
    )

    # Tracked parallel executions.
    parallel_executions = foo_invocation.execution[2]
    self.assertEqual(parallel_executions.id, 'agent@1:/a1/p1')
    self.assertIsInstance(parallel_executions, action_lib.ParallelExecutions)
    self.assertEqual(len(parallel_executions), 3)
    self.assertEqual(parallel_executions[0].id, 'agent@1:/a1/p1/b1')
    self.assertEqual(parallel_executions[1].id, 'agent@1:/a1/p1/b2')
    self.assertEqual(parallel_executions[2].id, 'agent@1:/a1/p1/b3')
    self.assertEqual(len(parallel_executions[0].queries), 1)
    self.assertEqual(len(parallel_executions[1].queries), 1)
    self.assertEqual(len(parallel_executions[2].queries), 1)

    # Invocation to Bar.
    bar_invocation = foo_invocation.execution[3]
    self.assertIs(bar_invocation.parent_action, foo_invocation)
    self.assertEqual(bar_invocation.id, 'agent@1:/a1/a1')
    self.assertIsInstance(bar_invocation, action_lib.ActionInvocation)
    self.assertIsInstance(bar_invocation.action, Bar)
    self.assertEqual(bar_invocation.result, 2)
    self.assertEqual(bar_invocation.metadata, dict(note='bar'))
    self.assertEqual(len(bar_invocation.execution.items), 2)

    # Save to HTML
    self.assertIn('result', session.to_html().content)

    # Save session to JSON
    json_str = session.to_json_str(save_ref_value=True)
    self.assertIsInstance(pg.from_json_str(json_str), action_lib.Session)

  def test_acall(self):
    bar = Bar()
    with lf.context(baz=1):
      r = bar.acall(lm=fake.StaticResponse('lm response'))
      self.assertEqual(asyncio.run(r), 3)

  def test_failed_action(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_action_error=True)
    with self.assertRaisesRegex(ValueError, 'Bar error'):
      foo(lm=lm)

    session = foo.session
    self.assertIsNotNone(session)
    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)

    # Inspecting the root invocation.
    root = session.root
    self.assertRegex(root.id, 'agent@.*:')
    self.assertTrue(root.has_error)
    foo_invocation = root.execution[0]
    self.assertIsInstance(foo_invocation, action_lib.ActionInvocation)
    self.assertTrue(foo_invocation.has_error)
    bar_invocation = foo_invocation.execution[3]
    self.assertIsInstance(bar_invocation, action_lib.ActionInvocation)
    self.assertTrue(bar_invocation.has_error)

    # Save to HTML
    self.assertIn('error', session.to_html().content)

  def test_failed_query(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_query_error=True)
    with self.assertRaisesRegex(lf_structured.MappingError, 'SyntaxError'):
      foo(lm=lm)

    session = foo.session
    self.assertIsNotNone(session)
    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)

    # Inspecting the root invocation.
    root = session.root
    self.assertRegex(root.id, 'agent@.*:')
    self.assertTrue(root.has_error)
    foo_invocation = root.execution[0]
    self.assertIsInstance(foo_invocation, action_lib.ActionInvocation)
    self.assertTrue(foo_invocation.has_error)
    self.assertEqual(len(foo_invocation.execution.items), 3)

  def test_succeeded_with_implicit_session(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1)
    foo(lm=lm, verbose=True)
    session = foo.session
    self.assertIsNotNone(session)
    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)
    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(session.final_result, 3)
    self.assertFalse(session.root.has_error)
    self.assertEqual(session.root.metadata, {})

  def test_failed_with_implicit_session(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_action_error=True)
    with self.assertRaisesRegex(ValueError, 'Bar error'):
      foo(lm=lm)
    session = foo.session
    self.assertIsNotNone(session)
    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)
    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertTrue(session.has_error)
    self.assertIsInstance(session.final_error, pg.ErrorInfo)
    self.assertIn('Bar error', str(session.root.error))

  def test_succeeded_with_explicit_session(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1)
    self.assertIsNone(foo.session)
    self.assertIsNone(foo.result)
    self.assertIsNone(foo.metadata)

    session = action_lib.Session(id='agent@1')
    self.assertEqual(session.id, 'agent@1')
    self.assertFalse(session.has_started)
    self.assertFalse(session.has_stopped)

    with session:
      result = foo(session, lm=lm, verbose=True)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(result, 3)
    self.assertIsNone(foo.session)
    self.assertEqual(foo.result, 3)
    self.assertEqual(
        foo.metadata, dict(note='foo', subtask_0=0, subtask_1=1, subtask_2=2)
    )
    self.assertIs(session.final_result, foo.result)
    self.assertFalse(session.has_error)

  def test_succeeded_with_explicit_session_start_end(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1)
    self.assertIsNone(foo.session)
    self.assertIsNone(foo.result)
    self.assertIsNone(foo.metadata)

    session = action_lib.Session(id='agent@1')
    self.assertEqual(session.id, 'agent@1')
    self.assertFalse(session.has_started)
    self.assertFalse(session.has_stopped)

    session.start()
    result = foo(session, lm=lm, verbose=True)
    session.end(result)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(result, 3)
    self.assertIsNone(foo.session)
    self.assertEqual(foo.result, 3)
    self.assertEqual(
        foo.metadata, dict(note='foo', subtask_0=0, subtask_1=1, subtask_2=2)
    )
    self.assertIs(session.final_result, foo.result)
    self.assertFalse(session.has_error)

  def test_failed_with_explicit_session(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_action_error=True)
    session = action_lib.Session(id='agent@1')
    with self.assertRaisesRegex(ValueError, 'Bar error'):
      with session:
        foo(session, lm=lm, verbose=True)
    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertTrue(session.has_error)
    self.assertIsNone(session.final_result)
    self.assertIsInstance(session.root.error, pg.ErrorInfo)
    self.assertIn('Bar error', str(session.root.error))

  def test_failed_with_explicit_session_without_start(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_action_error=True)
    session = action_lib.Session(id='agent@1')
    with self.assertRaisesRegex(ValueError, 'Please call `Session.start'):
      foo(session, lm=lm, verbose=True)

  def test_succeed_with_multiple_actions(self):
    lm = fake.StaticResponse('lm response')
    with action_lib.Session() as session:
      x = Bar()(session, lm=lm)
      y = Bar()(session, lm=lm)
      self.assertTrue(session.has_started)
      self.assertFalse(session.has_stopped)
      session.add_metadata(note='root metadata')
      session.end(x + y)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(session.final_result, 2 + 2)
    self.assertEqual(len(session.root.execution), 2)
    self.assertEqual(session.root.metadata, dict(note='root metadata'))

  def test_failed_with_multiple_actions(self):
    lm = fake.StaticResponse('lm response')
    with self.assertRaisesRegex(ValueError, 'Bar error'):
      with action_lib.Session() as session:
        x = Bar()(session, lm=lm)
        y = Bar(simulate_action_error=True)(session, lm=lm)
        session.end(x + y)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertTrue(session.has_error)
    self.assertIsInstance(session.root.error, pg.ErrorInfo)
    self.assertEqual(len(session.root.execution), 3)
    self.assertEqual(len(session.root.actions), 2)
    self.assertEqual(len(session.root.logs), 1)
    self.assertFalse(session.root.execution[0].has_error)
    self.assertTrue(session.root.execution[1].has_error)

  def test_max_execution_time(self):
    lm = fake.StaticResponse('lm response')
    bar = Bar(simulate_execution_time=1)
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Bar.*has exceeded .* 0.5 seconds'
    ):
      bar(lm=lm, max_execution_time=0.5)

    foo = Foo(1, simulate_execution_time=[0, 0, 0, 1])
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Foo.* has exceeded .* 0.5 seconds'
    ):
      foo(lm=lm, max_execution_time=0.5)

    # Timeout within concurrent_map.
    foo = Foo(1, simulate_execution_time=[0, 0, 1, 0])
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Foo.* has exceeded .* 0.5 seconds'
    ):
      foo(lm=lm, max_execution_time=0.5)

    # Timeout within bar.
    foo = Foo(
        1, simulate_execution_time=[0, 0, 0, 1], max_bar_execution_time=0.5
    )
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Bar.* has exceeded .* 0.5 seconds'
    ):
      foo(lm=lm)

    # Timeout within bar, however the effective max_execution_time of bar is the
    # remaining time of the parent action as it's smaller (0.5 < 1).
    foo = Foo(
        1, simulate_execution_time=[0, 0.5, 0, 1.0], max_bar_execution_time=1.0
    )
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Foo.* has exceeded .*1.0 seconds'
    ):
      foo(lm=lm, max_execution_time=1.0)

  def test_log(self):
    session = action_lib.Session()
    session.debug('hi', x=1, y=2)
    session.info('hi', x=1, y=2, for_action=session.root)
    session.warning('hi', x=1, y=2, for_action=session.root.action)
    session.error('hi', x=1, y=2)
    session.fatal('hi', x=1, y=2)

  def test_as_message(self):
    session = action_lib.Session()
    self.assertIn('agent@', session.id)
    self.assertIsInstance(session.as_message(), lf.AIMessage)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Vertex AI generative models."""

import datetime
import functools
import os
from typing import Annotated, Any, Literal

import langfun.core as lf
from langfun.core.llms import anthropic
from langfun.core.llms import gemini
from langfun.core.llms import openai_compatible
from langfun.core.llms import rest
import pyglove as pg

try:
  # pylint: disable=g-import-not-at-top
  from google import auth as google_auth
  from google.auth import exceptions as auth_exceptions
  from google.auth import credentials as credentials_lib
  from google.auth.transport import requests as auth_requests
  # pylint: enable=g-import-not-at-top

  Credentials = credentials_lib.Credentials
except ImportError:
  google_auth = None
  auth_exceptions = None
  credentials_lib = None
  auth_requests = None
  Credentials = Any


@pg.use_init_args(['api_endpoint'])
class VertexAI(rest.REST):
  """Base class for VertexAI models.

  This class handles the authentication of vertex AI models. Subclasses
  should implement `request` and `result` methods, as well as the `api_endpoint`
  property. Or let users to provide them as __init__ arguments.

  Please check out VertexAIGemini in `gemini.py` as an example.
  """

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE,
          [
              m.model_id for m in gemini.SUPPORTED_MODELS
              if m.provider == 'VertexAI' or (
                  isinstance(m.provider, pg.hyper.OneOf)
                  and 'VertexAI' in m.provider.candidates
              )
          ]
      ),
      'The name of the model to use.',
  ]

  model: Annotated[
      str | None,
      'Model name.'
  ] = None

  project: Annotated[
      str | None,
      (
          'Vertex AI project ID. Or set from environment variable '
          'VERTEXAI_PROJECT.'
      ),
  ] = None

  location: Annotated[
      str | None,
      (
          'Vertex AI service location. Or set from environment variable '
          'VERTEXAI_LOCATION.'
      ),
  ] = None

  credentials: Annotated[
      Credentials | None,
      (
          'Credentials to use. If None, the default credentials to the '
          'environment will be used.'
      ),
  ] = None

  def _on_bound(self):
    super()._on_bound()
    if google_auth is None:
      raise ValueError(
          'Please install "langfun[llm-google-vertex]" to use Vertex AI models.'
      )
    self._project = None
    self._credentials = None

  def _initialize(self):
    project = self.project or os.environ.get('VERTEXAI_PROJECT', None)
    if not project:
      raise ValueError(
          'Please specify `project` during `__init__` or set environment '
          'variable `VERTEXAI_PROJECT` with your Vertex AI project ID.'
      )

    location = self.location or os.environ.get('VERTEXAI_LOCATION', None)
    if not location:
      raise ValueError(
          'Please specify `location` during `__init__` or set environment '
          'variable `VERTEXAI_LOCATION` with your Vertex AI service location.'
      )

    self._project = project
    self._location = location

    credentials = self.credentials
    if credentials is None:
      # Use default credentials.
      credentials, _ = google_auth.default(
          scopes=['https://www.googleapis.com/auth/cloud-platform']
      )
    self._credentials = credentials

  def session(self):
    assert self._api_initialized
    assert self._credentials is not None
    assert auth_requests is not None
    s = auth_requests.AuthorizedSession(self._credentials)
    s.headers.update(self.headers or {})
    return s

  def _sample_single(self, prompt: lf.Message) -> lf.LMSamplingResult:
    assert auth_exceptions is not None
    try:
      return super()._sample_single(prompt)
    except (
        auth_exceptions.RefreshError,
    ) as e:
      raise lf.TemporaryLMError(
          f'Failed to refresh Google authentication credentials: {e}'
      ) from e

#
# Gemini models served by Vertex AI.
#


@pg.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class VertexAIGemini(VertexAI, gemini.Gemini):
  """Gemini models served by Vertex AI.."""

  # Set default location to us-central1.
  location = 'us-central1'

  @property
  def api_endpoint(self) -> str:
    assert self._api_initialized
    return (
        f'https://aiplatform.googleapis.com/v1/projects/'
        f'{self._project}/locations/{self._location}/publishers/google/'
        f'models/{self.model}:generateContent'
    )

  @functools.cached_property
  def model_info(self) -> gemini.GeminiModelInfo:
    return super().model_info.clone(override=dict(provider='VertexAI'))


#
# Production models.
#
class VertexAIGemini25Pro(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.5 Pro GA model launched on 06/17/2025."""

  model = 'gemini-2.5-pro'
  location = 'global'


class VertexAIGemini25Flash(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.5 Flash GA model launched on 06/17/2025."""

  model = 'gemini-2.5-flash'
  location = 'global'


class VertexAIGemini25ProPreview_20250605(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.5 Pro model launched on 06/05/2025."""
  model = 'gemini-2.5-pro-preview-06-05'
  location = 'global'


class VertexAIGemini25FlashPreview_20250520(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.5 Flash model launched on 05/20/2025."""
  model = 'gemini-2.5-flash-preview-05-20'


class VertexAIGemini25ProPreview_20250506(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.5 Pro model launched on 05/06/2025."""
  model = 'gemini-2.5-pro-preview-05-06'


class VertexAIGemini25FlashPreview_20250417(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.5 Flash model launched on 04/17/2025."""
  model = 'gemini-2.5-flash-preview-04-17'


class VertexAIGemini25ProPreview_20250325(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.5 Pro model launched on 03/25/2025."""
  model = 'gemini-2.5-pro-preview-03-25'


class VertexAIGemini25ProExp_20250325(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.5 Pro model launched on 03/25/2025."""
  model = 'gemini-2.5-pro-exp-03-25'


class VertexAIGemini2Flash(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 model (latest stable)."""
  model = 'gemini-2.0-flash'


class VertexAIGemini2Flash_001(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 model version 001."""
  model = 'gemini-2.0-flash-001'


class VertexAIGemini2FlashLitePreview_20250205(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.0 Flash lite preview model launched on 02/05/2025."""
  model = 'gemini-2.0-flash-lite-preview-02-05'


class VertexAIGemini15Pro(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model (latest stable)."""
  model = 'gemini-1.5-pro'


class VertexAIGemini15Pro_002(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model (version 002)."""
  model = 'gemini-1.5-pro-002'


class VertexAIGemini15Pro_001(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model (version 001)."""
  model = 'gemini-1.5-pro-001'


class VertexAIGemini15Flash(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model (latest stable)."""
  model = 'gemini-1.5-flash'


class VertexAIGemini15Flash_002(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model (version 002)."""
  model = 'gemini-1.5-flash-002'


class VertexAIGemini15Flash_001(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model (version 001)."""
  model = 'gemini-1.5-flash-001'


class VertexAIGemini15Flash8B(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash 8B model (latest stable)."""
  model = 'gemini-1.5-flash-8b'


class VertexAIGemini15Flash8B_001(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash 8B model (version 001)."""
  model = 'gemini-1.5-flash-8b-001'

#
# Experimental models.
#


class VertexAIGemini25FlashImagePreview(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini 2.5 Flash Image Preview model."""

  model = 'gemini-2.5-flash-image-preview'
  location = 'global'
  response_modalities = ['TEXT', 'IMAGE']


class VertexAIGemini2ProExp_20250205(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 Pro model launched on 02/05/2025."""
  model = 'gemini-2.0-pro-exp-02-05'


class VertexAIGemini2FlashThinkingExp_20250121(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 Thinking model launched on 01/21/2025."""
  model = 'gemini-2.0-flash-thinking-exp-01-21'
  timeout = None


class VertexAIGeminiFlash2_0Exp(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 2.0 Flash model."""
  model = 'gemini-2.0-flash-exp'


class VertexAIGeminiExp_20241206(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini Experimental model launched on 12/06/2024."""
  model = 'gemini-exp-1206'


#
# Anthropic models on Vertex AI.
#


@pg.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class VertexAIAnthropic(VertexAI, anthropic.Anthropic):
  """Anthropic models on VertexAI."""

  location: Annotated[
      Literal['us-east5', 'europe-west1'],
      'GCP location with Anthropic models hosted.'
  ] = 'us-east5'

  api_version = 'vertex-2023-10-16'

  @functools.cached_property
  def model_info(self) -> lf.ModelInfo:
    mi = anthropic._SUPPORTED_MODELS_BY_MODEL_ID[self.model]  # pylint: disable=protected-access
    if mi.provider != 'VertexAI':
      for m in anthropic.SUPPORTED_MODELS:
        if m.provider == 'VertexAI' and m.alias_for == m.model_id:
          mi = m
          self.rebind(model=mi.model_id, skip_notification=True)
          break
    return mi

  @property
  def headers(self):
    return {
        'Content-Type': 'application/json; charset=utf-8',
    }

  @property
  def api_endpoint(self) -> str:
    return (
        f'https://{self.location}-aiplatform.googleapis.com/v1/projects/'
        f'{self._project}/locations/{self.location}/publishers/anthropic/'
        f'models/{self.model}:streamRawPredict'
    )

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ):
    request = super().request(prompt, sampling_options)
    request['anthropic_version'] = self.api_version
    del request['model']
    return request


# pylint: disable=invalid-name


class VertexAIClaude4Opus_20250514(VertexAIAnthropic):
  """Anthropic's Claude 4 Opus model on VertexAI."""
  model = 'claude-opus-4@20250514'
  location = 'us-east5'


class VertexAIClaude4Sonnet_20250514(VertexAIAnthropic):
  """Anthropic's Claude 4 Sonnet model on VertexAI."""
  model = 'claude-sonnet-4@20250514'
  location = 'us-east5'


class VertexAIClaude37Sonnet_20250219(VertexAIAnthropic):
  """Anthropic's Claude 3.7 model on VertexAI."""
  model = 'claude-3-7-sonnet@20250219'
  location = 'us-east5'


class VertexAIClaude35Sonnet_20241022(VertexAIAnthropic):
  """Anthropic's Claude 3.5 Sonnet model on VertexAI."""
  model = 'claude-3-5-sonnet-v2@20241022'


class VertexAIClaude35Haiku_20241022(VertexAIAnthropic):
  """Anthropic's Claude 3.5 Haiku model on VertexAI."""
  model = 'claude-3-5-haiku@20241022'


class VertexAIClaude3Opus_20240229(VertexAIAnthropic):
  """Anthropic's Claude 3 Opus model on VertexAI."""
  model = 'claude-3-opus@20240229'

# pylint: enable=invalid-name

#
# Llama models on Vertex AI.
# https://cloud.google.com/vertex-ai/generative-ai/docs/partner-models/llama
# Pricing: https://cloud.google.com/vertex-ai/generative-ai/pricing#meta-models

LLAMA_MODELS = [
    lf.ModelInfo(
        model_id='llama-3.1-405b-instruct-maas',
        in_service=True,
        model_type='instruction-tuned',
        provider='VertexAI',
        description=(
            'Llama 3.2 405B vision instruct model on VertexAI (Preview)'
        ),
        url='https://huggingface.co/meta-llama/Llama-3.1-405B',
        release_date=datetime.datetime(2024, 7, 23),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=5.0,
            cost_per_1m_output_tokens=16.0,
        ),
        rate_limits=None,
    ),
    lf.ModelInfo(
        model_id='llama-3.2-90b-vision-instruct-maas',
        in_service=True,
        model_type='instruction-tuned',
        provider='VertexAI',
        description=(
            'Llama 3.2 90B vision instruct model on VertexAI (Preview)'
        ),
        release_date=datetime.datetime(2024, 7, 23),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=8_192,
        ),
        # Free during preview.
        pricing=None,
        rate_limits=None,
    ),
    lf.ModelInfo(
        model_id='llama-3.1-70b-instruct-maas',
        in_service=True,
        model_type='instruction-tuned',
        provider='VertexAI',
        description=(
            'Llama 3.2 70B vision instruct model on VertexAI (Preview)'
        ),
        release_date=datetime.datetime(2024, 7, 23),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=8_192,
        ),
        # Free during preview.
        pricing=None,
        rate_limits=None,
    ),
    lf.ModelInfo(
        model_id='llama-3.1-8b-instruct-maas',
        in_service=True,
        model_type='instruction-tuned',
        provider='VertexAI',
        description=(
            'Llama 3.2 8B vision instruct model on VertexAI (Preview)'
        ),
        release_date=datetime.datetime(2024, 7, 23),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=0,
            max_output_tokens=0,
        ),
        # Free during preview.
        pricing=None,
        rate_limits=None,
    ),
]

_LLAMA_MODELS_BY_MODEL_ID = {m.model_id: m for m in LLAMA_MODELS}


@pg.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class VertexAILlama(VertexAI, openai_compatible.OpenAICompatible):
  """Llama models on VertexAI."""

  model: pg.typing.Annotated[
      pg.typing.Enum(pg.MISSING_VALUE, [m.model_id for m in LLAMA_MODELS]),
      'Llama model ID.',
  ]

  locations: Annotated[
      Literal['us-central1'],
      (
          'GCP locations with Llama models hosted. '
          'See https://cloud.google.com/vertex-ai/generative-ai/docs/partner-models/llama#regions-quotas'
      )
  ] = 'us-central1'

  @functools.cached_property
  def model_info(self) -> lf.ModelInfo:
    return _LLAMA_MODELS_BY_MODEL_ID[self.model]

  @property
  def api_endpoint(self) -> str:
    assert self._api_initialized
    return (
        f'https://{self._location}-aiplatform.googleapis.com/v1beta1/projects/'
        f'{self._project}/locations/{self._location}/endpoints/'
        f'openapi/chat/completions'
    )

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ):
    request = super().request(prompt, sampling_options)
    request['model'] = f'meta/{self.model}'
    return request


# pylint: disable=invalid-name
class VertexAILlama32_90B(VertexAILlama):
  """Llama 3.2 90B vision instruct model on VertexAI."""
  model = 'llama-3.2-90b-vision-instruct-maas'


class VertexAILlama31_405B(VertexAILlama):
  """Llama 3.1 405B vision instruct model on VertexAI."""
  model = 'llama-3.1-405b-instruct-maas'


class VertexAILlama31_70B(VertexAILlama):
  """Llama 3.1 70B vision instruct model on VertexAI."""
  model = 'llama-3.1-70b-instruct-maas'


class VertexAILlama31_8B(VertexAILlama):
  """Llama 3.1 8B vision instruct model on VertexAI."""
  model = 'llama-3.1-8b-instruct-maas'


# pylint: enable=invalid-name

#
# Mistral models on Vertex AI.
# pylint: disable=line-too-long
# Models: https://cloud.google.com/vertex-ai/generative-ai/docs/partner-models/mistral
# Pricing: https://cloud.google.com/vertex-ai/generative-ai/pricing#mistral-models
# pylint: enable=line-too-long

MISTRAL_MODELS = [
    lf.ModelInfo(
        model_id='mistral-large-2411',
        in_service=True,
        model_type='instruction-tuned',
        provider='VertexAI',
        description='Mistral Large model on VertexAI (GA) version 11/21/2024',
        release_date=datetime.datetime(2024, 11, 21),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=2.0,
            cost_per_1m_output_tokens=6.0,
        ),
        rate_limits=None,
    ),
    lf.ModelInfo(
        model_id='codestral-2501',
        in_service=True,
        model_type='instruction-tuned',
        provider='VertexAI',
        description=(
            'Mistral Codestral model on VertexAI (GA) (version 01/13/2025)'
        ),
        release_date=datetime.datetime(2025, 1, 13),
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=128_000,
            max_output_tokens=8_192,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_input_tokens=0.3,
            cost_per_1m_output_tokens=0.9,
        ),
        rate_limits=None,
    ),
]

_MISTRAL_MODELS_BY_MODEL_ID = {m.model_id: m for m in MISTRAL_MODELS}


@pg.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class VertexAIMistral(VertexAI, openai_compatible.OpenAICompatible):
  """Mistral AI models on VertexAI."""

  model: pg.typing.Annotated[
      pg.typing.Enum(pg.MISSING_VALUE, [m.model_id for m in MISTRAL_MODELS]),
      'Mistral model ID.',
  ]

  locations: Annotated[
      Literal['us-central1', 'europe-west4'],
      (
          'GCP locations with Mistral models hosted. '
          'See https://cloud.google.com/vertex-ai/generative-ai/docs/partner-models/mistral#regions-quotas'
      )
  ] = 'us-central1'

  @functools.cached_property
  def model_info(self) -> lf.ModelInfo:
    return _MISTRAL_MODELS_BY_MODEL_ID[self.model]

  @property
  def api_endpoint(self) -> str:
    assert self._api_initialized
    return (
        f'https://{self._location}-aiplatform.googleapis.com/v1/projects/'
        f'{self._project}/locations/{self._location}/publishers/mistralai/'
        f'models/{self.model}:rawPredict'
    )


# pylint: disable=invalid-name
class VertexAIMistralLarge_20241121(VertexAIMistral):
  """Mistral Large model on VertexAI released on 2024/11/21."""
  model = 'mistral-large-2411'


class VertexAICodestral_20250113(VertexAIMistral):
  """Mistral Nemo model on VertexAI released on 2024/07/24."""
  model = 'codestral-2501'

# pylint: enable=invalid-name


#
# Register Vertex AI models so they can be retrieved with LanguageModel.get().
#


def _register_vertexai_models():
  """Register Vertex AI models."""
  for m in gemini.SUPPORTED_MODELS:
    if m.provider == 'VertexAI' or (
        isinstance(m.provider, pg.hyper.OneOf)
        and 'VertexAI' in m.provider.candidates
    ):
      lf.LanguageModel.register(m.model_id, VertexAIGemini)

  for m in anthropic.SUPPORTED_MODELS:
    if m.provider == 'VertexAI':
      lf.LanguageModel.register(m.model_id, VertexAIAnthropic)

  for m in LLAMA_MODELS:
    lf.LanguageModel.register(m.model_id, VertexAILlama)

  for m in MISTRAL_MODELS:
    lf.LanguageModel.register(m.model_id, VertexAIMistral)

_register_vertexai_models()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import unittest
from langfun.core.llms import azure_openai


class AzureOpenAITest(unittest.TestCase):
  def test_api_key_missing(self):
    # Ensure that ValueError is raised when API key is not provided
    os.environ.pop('AZURE_OPENAI_API_KEY', None)
    with self.assertRaisesRegex(ValueError, 'Azure OpenAI requires an API key'):
      azure = azure_openai.AzureOpenAI(model='gpt-4')
      azure._initialize()

  def test_api_endpoint(self):
    # Test that api_endpoint is properly constructed
    azure = azure_openai.AzureOpenAI(model='gpt-4', api_key='test_key')
    azure._initialize()
    expected = (
        'https://api.openai.azure.com/openai/deployments/gpt-4/chat/'
        'completions?api-version=2023-05-15'
    )
    self.assertEqual(azure.api_endpoint, expected)

  def test_headers(self):
    # Test that headers contain the api-key with correct value
    azure = azure_openai.AzureOpenAI(model='gpt-4', api_key='test_key')
    azure._initialize()
    headers = azure.headers
    self.assertIn('api-key', headers)
    self.assertEqual(headers.get('api-key'), 'test_key')


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic parsing."""

from typing import Any, Callable, Type, Union

import langfun.core as lf
from langfun.core.structured import mapping
from langfun.core.structured import querying
from langfun.core.structured import schema as schema_lib
import pyglove as pg


@lf.use_init_args(['schema', 'default', 'examples'])
class _ParseStructure(mapping.Mapping):
  """Parse an object out from a natural language text."""

  context_title = 'USER_REQUEST'
  input_title = 'LM_RESPONSE'

  # Require input to be message.
  input: lf.Message = lf.contextual()

  # Mark schema as required.
  schema: pg.typing.Annotated[
      schema_lib.schema_spec(), 'Required schema for parsing.'
  ]


class _ParseStructureJson(_ParseStructure):
  """Parse an object out from a NL text using JSON as the protocol."""

  preamble = """
      Please help translate the last LM response into JSON based on the request and the schema:

      INSTRUCTIONS:
        1. If the schema has `_type`, carry it over to the JSON output.
        2. If a field from the schema cannot be extracted from the response, use null as the JSON value.
      """

  protocol = 'json'
  schema_title = 'SCHEMA'
  output_title = 'JSON'


class _ParseStructurePython(_ParseStructure):
  """Parse an object out from a NL text using Python as the protocol."""

  preamble = """
      Please help translate the last {{ input_title }} into {{ output_title}} based on {{ schema_title }}.

      INSTRUCTIONS:
      1. Both {{ schema_title }} and {{ output_title }} are described in Python.
      2. {{ output_title }} must be created solely based on the information provided in {{ input_title }}.
      """

  protocol = 'python'
  schema_title = 'RESULT_TYPE'
  output_title = 'RESULT_OBJECT'


def parse(
    message: Union[lf.Message, str],
    schema: Union[
        schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any]
    ],
    default: Any = lf.RAISE_IF_HAS_ERROR,
    *,
    user_prompt: str | None = None,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    include_context: bool = False,
    cache_seed: int | None = 0,
    autofix: int = 0,
    autofix_lm: lf.LanguageModel | None = None,
    protocol: schema_lib.SchemaProtocol = 'python',
    returns_message: bool = False,
    **kwargs,
) -> Any:
  """Parse a natural language message based on schema.

  Examples:

    ```
    class FlightDuration(pg.Object):
      hours: int
      minutes: int

    class Flight(pg.Object):
      airline: str
      flight_number: str
      departure_airport_code: str
      arrival_airport_code: str
      departure_time: str
      arrival_time: str
      duration: FlightDuration
      stops: int
      price: float

    input = '''
      The flight is operated by United Airlines, has the flight number UA2631,
      departs from San Francisco International Airport (SFO), arrives at John
      F. Kennedy International Airport (JFK), It departs at 2023-09-07T05:15:00,
      arrives at 2023-09-07T12:12:00, has a duration of 7 hours and 57 minutes,
      makes 1 stop, and costs $227.
      '''

    r = lf.parse(input, Flight)
    assert isinstance(r, Flight)
    assert r.airline == 'United Airlines'
    assert r.departure_airport_code == 'SFO'
    assert r.duration.hour = 7
    ```

  Args:
    message: A `lf.Message` object  or a string as the natural language input.
      It provides the complete context for the parsing.
    schema: A `lf.transforms.ParsingSchema` object or equivalent annotations.
    default: The default value if parsing failed. If not specified, error will
      be raised.
    user_prompt: An optional user prompt as the description or ask for the
      message, which provide more context for parsing.
    lm: The language model to use. If not specified, the language model from
      `lf.context` context manager will be used.
    examples: An optional list of fewshot examples for helping parsing. If None,
      the default one-shot example will be added.
    include_context: If True, include the request sent to LLM for obtaining the
      response to pares. Otherwise include only the response.
    cache_seed: Seed for computing cache key. The cache key is determined by a
      tuple of (lm, prompt, cache seed). If None, cache will be disabled for
      the query even cache is configured by the LM.
    autofix: Number of attempts to auto fix the generated code. If 0, autofix is
      disabled. Auto-fix is not supported for 'json' protocol.
    autofix_lm: The language model to use for autofix. If not specified, the
      `autofix_lm` from `lf.context` context manager will be used. Otherwise it
      will use `lm`.
    protocol: The protocol for schema/value representation. Applicable values
      are 'json' and 'python'. By default 'python' will be used.`
    returns_message: If True, returns `lf.Message` as the output, instead of
      returning the structured `message.result`.
    **kwargs: Keyword arguments passed to the `lf.structured.ParseStructure`
      transform.

  Returns:
    The parsed result based on the schema.
  """
  # Autofix is not supported for JSON yet.
  if protocol == 'json':
    autofix = 0

  message = lf.AIMessage.from_value(message)
  if message.source is None and user_prompt is not None:
    message.source = lf.UserMessage(user_prompt, tags=['lm-input'])
  context = getattr(message.lm_input, 'text', None) if include_context else None

  t = _parse_structure_cls(protocol)(
      schema=schema,
      context=context,
      default=default,
      examples=examples or default_parse_examples(),
  )

  # Setting up context.
  call_context = dict(cache_seed=cache_seed, autofix=autofix)
  if lm is not None:
    call_context['lm'] = lm
  autofix_lm = autofix_lm or lm
  if autofix_lm is not None:
    call_context['autofix_lm'] = autofix_lm
  call_context.update(kwargs)

  output = t(input=message, **call_context)
  return output if returns_message else output.result


async def aparse(
    message: Union[lf.Message, str],
    schema: Union[
        schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any]
    ],
    default: Any = lf.RAISE_IF_HAS_ERROR,
    *,
    user_prompt: str | None = None,
    lm: lf.LanguageModel | None = None,
    examples: list[mapping.MappingExample] | None = None,
    include_context: bool = False,
    cache_seed: int | None = 0,
    autofix: int = 0,
    autofix_lm: lf.LanguageModel | None = None,
    protocol: schema_lib.SchemaProtocol = 'python',
    returns_message: bool = False,
    **kwargs,
) -> Any:
  """Async version of `lf.parse`."""
  # TODO(daiyip): implement native async parsing.
  return await lf.invoke_async(
      parse,
      message,
      schema,
      default,
      user_prompt=user_prompt,
      lm=lm,
      examples=examples,
      include_context=include_context,
      cache_seed=cache_seed,
      autofix=autofix,
      autofix_lm=autofix_lm,
      protocol=protocol,
      returns_message=returns_message,
      **kwargs
  )


def call(
    prompt: str | lf.Template,
    schema: Union[
        None, schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any]
    ] = None,
    *,
    lm: lf.LanguageModel | None = None,
    parsing_lm: lf.LanguageModel | None = None,
    parsing_examples: list[mapping.MappingExample] | None = None,
    parsing_include_context: bool = False,
    cache_seed: int | None = 0,
    autofix: int = 0,
    autofix_lm: lf.LanguageModel | None = None,
    response_postprocess: Callable[[str], str] | None = None,
    protocol: schema_lib.SchemaProtocol = 'python',
    returns_message: bool = False,
    **kwargs,
) -> Any:
  """Call a language model with prompt and formulate response in return type.

  Examples::

    # Call with constant string-type prompt.
    lf.call('Compute one plus one', lm=lf.llms.Gpt35())
    >> "two"

    # Call with returning a structured (int) type.
    lf.call('Compute one plus one', int, lm=lf.llms.Gpt35())
    >> 2

    # Call with a template string with variables.
    lf.call('Compute {{x}} plus {{y}}', int,
            x='one', y='one', lm=lf.llms.Gpt35())
    >> 2

    # Call with an `lf.Template` object with variables.
    lf.call(lf.Template('Compute {{x}} plus {{y}}', x=1), int,
            y=1, lm=lf.llms.Gpt35())
    >> 2

  Args:
    prompt: User prompt that will be sent to LM, which could be a string or a
      string template whose variables are provided from **kwargs.
    schema: Type annotations for return type. If None, the raw LM response will
      be returned (str). Otherwise, the response will be parsed based on the
      return type.
    lm: Language model used to prompt to get a natural language-based response.
      If not specified, `lm` from `lf.context` context manager will be used.
    parsing_lm: Language model that will be used for parsing. If None, the `lm`
      for prompting the LM will be used.
    parsing_examples: Examples for parsing the output. If None,
      `lf.structured.DEFAULT_PARSE_EXAMPLES` will be used.
    parsing_include_context: If True, include the request sent to LLM for
      obtaining the response to pares. Otherwise include only the response.
    cache_seed: Seed for computing cache key. The cache key is determined by a
      tuple of (lm, prompt, cache seed). If None, cache will be disabled for
      the query even cache is configured by the LM.
    autofix: Number of attempts to auto fix the generated code. If 0, autofix is
      disabled. Auto-fix is not supported for 'json' protocol.
    autofix_lm: The language model to use for autofix. If not specified, the
      `autofix_lm` from `lf.context` context manager will be used. Otherwise it
      will use `parsing_lm`.
    response_postprocess: A callback function to post process the text response
      before sending for parsing.
    protocol: The protocol for schema/value representation. Applicable values
      are 'json' and 'python'. By default 'python' will be used.`
    returns_message: If True, return a `lf.Message` object instead of its text
      or result.
    **kwargs: Keyword arguments. Including options that control the calling
      behavior, such as `lm`, `temperature`, etc. As well as variables that will
      be fed to the prompt if it's a string template.

  Returns:
    A string if `returns` is None or an instance of the return type.
  """
  # Call `lm` for natural response.
  lm_output = lf.LangFunc.from_value(prompt, **kwargs)(lm=lm)

  if response_postprocess is not None:
    postprocessed_text = response_postprocess(lm_output.text)
    if postprocessed_text != lm_output.text:
      processed_lm_output = lf.AIMessage(postprocessed_text, source=lm_output)
      lm_output = processed_lm_output

  if schema in (str, None):
    return lm_output if returns_message else lm_output.text

  def _chain_nl_output_message(parsing_message: lf.Message):
    """Chain the source of the parsed output to the LM output."""
    parsing_message.root.source = lm_output
    parsing_message.tag('parsing-lm-output')
    parsing_message.lm_input.tag('parsing-lm-input')

  # Call `parsing_lm` for structured parsing.
  try:
    parsing_message = querying.query(
        lm_output.text,
        schema,
        examples=parsing_examples,
        lm=parsing_lm or lm,
        include_context=parsing_include_context,
        cache_seed=cache_seed,
        autofix=autofix,
        autofix_lm=autofix_lm or lm,
        protocol=protocol,
        returns_message=True,
        **kwargs,
    )
    _chain_nl_output_message(parsing_message)
  except mapping.MappingError as e:
    _chain_nl_output_message(e.lm_response)
    raise e
  return parsing_message if returns_message else parsing_message.result


async def acall(
    prompt: str | lf.Template,
    schema: Union[
        None, schema_lib.Schema, Type[Any], list[Type[Any]], dict[str, Any]
    ] = None,
    *,
    lm: lf.LanguageModel | None = None,
    parsing_lm: lf.LanguageModel | None = None,
    parsing_examples: list[mapping.MappingExample] | None = None,
    parsing_include_context: bool = False,
    cache_seed: int | None = 0,
    autofix: int = 0,
    autofix_lm: lf.LanguageModel | None = None,
    response_postprocess: Callable[[str], str] | None = None,
    protocol: schema_lib.SchemaProtocol = 'python',
    returns_message: bool = False,
    **kwargs,
) -> Any:
  """Async version of `lf.call`."""
  # TODO(daiyip): implement native async calling.
  return await lf.invoke_async(
      call,
      prompt,
      schema,
      lm=lm,
      parsing_lm=parsing_lm,
      parsing_examples=parsing_examples,
      parsing_include_context=parsing_include_context,
      cache_seed=cache_seed,
      autofix=autofix,
      autofix_lm=autofix_lm,
      response_postprocess=response_postprocess,
      protocol=protocol,
      returns_message=returns_message,
      **kwargs
  )


def _parse_structure_cls(
    protocol: schema_lib.SchemaProtocol,
) -> Type[_ParseStructure]:
  if protocol == 'json':
    return _ParseStructureJson
  elif protocol == 'python':
    return _ParseStructurePython
  else:
    raise ValueError(f'Unknown protocol: {protocol!r}.')


def default_parse_examples() -> list[mapping.MappingExample]:
  """Default parsing examples."""

  class AdditionResults(pg.Object):
    one_plus_one_equals: int | None
    two_plus_two_equals: int | None

  return [
      mapping.MappingExample(
          input='Two plus two equals four. Three plus three equals six.',
          schema=AdditionResults,
          output=AdditionResults(
              one_plus_one_equals=None, two_plus_two_equals=4
          ),
      ),
  ]



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python code error correction."""
from typing import Any
import langfun.core as lf
from langfun.core.coding.python import execution
import pyglove as pg


class CodeWithError(pg.Object):
  """Python code with error."""

  code: str
  error: str


class CorrectedCode(pg.Object):
  corrected_code: str


def run_with_correction(
    code: str,
    error: str | None = None,
    *,
    global_vars: dict[str, Any] | None = None,
    lm: lf.LanguageModel | None = None,
    max_attempts: int = 5,
    sandbox: bool | None = None,
    permission: pg.coding.CodePermission = pg.coding.CodePermission.ALL,
    timeout: int | None = 5,
    returns_code: bool = False,
    returns_stdout: bool = False,
    outputs_intermediate: bool = False,
) -> Any | tuple[Any, str]:
  """Correct code with a language model via self-play.

  Args:
    code: The source code that may or may not be problematic.
    error: An optional initial error for `code` when it's problematic, usually
      caught from elsewhere when it ran. If None, code will be executed once to
      verify if its good and obtain a feedback error message.
    global_vars: A dict of str to value as the global variables that could be
      accessed within the corrected code.
    lm: Language model to be used. If not specified, it will try to use the `lm`
      under `lf.context`.
    max_attempts: Max number of attempts for the correction.
    sandbox: If True, run code in sandbox; If False, run code in current
      process. If None, run in sandbox first, if the output could not be
      serialized and pass to current process, run the code again in current
      process.
    permission: The permission to run the code.
    timeout: The timeout for running the corrected code. If None, there is no
      timeout. Applicable only when sandbox is set to True.
    returns_code: If True, the return value is a tuple of (result, final code).
      Otherwise the return value is the result only.
    returns_stdout: If True, the stdout (a str) will be returned.
    outputs_intermediate: If True, intermediate output will be outputted as a
      dict, with the last line's value accessible by key '__result__'. Otherwise
      the value of the last line will be returned.

  Returns:
    Run result if `returns_code` is set to False (default), otherwise a tuple
      of (result, final code str).

  Raises:
    `lf.CodeError`: If code cannot be corrected after `max_attempts`.
  """
  # Delay import at runtime to avoid circular depenency.
  # pylint: disable=g-import-not-at-top
  # pytype: disable=import-error
  from langfun.core.structured import querying
  # pytype: enable=import-error
  # pylint: enable=g-import-not-at-top

  if max_attempts == 0:
    result = _maybe_custom_validate(
        execution.run(
            code,
            global_vars=global_vars,
            sandbox=sandbox,
            timeout=timeout,
            permission=permission,
            returns_stdout=returns_stdout,
            outputs_intermediate=outputs_intermediate,
        )
    )
    return (result, code) if returns_code else result

  def result_and_error(code: str) -> tuple[Any, str | None]:
    try:
      result = _maybe_custom_validate(
          execution.run(
              code,
              global_vars=global_vars,
              sandbox=sandbox,
              timeout=timeout,
              permission=permission,
              outputs_intermediate=outputs_intermediate,
          )
      )
      return (result, None)
    except Exception as e:  # pylint: disable=broad-exception-caught
      return (None, _error_feedback_str(e))

  if error is None:
    result, error = result_and_error(code)
    if error is None:
      return (result, code) if returns_code else result

  num_attempts = 0
  for _ in range(max_attempts):
    num_attempts += 1
    # Catch query() error when the LM could not formulate a completion
    # structure.
    try:
      # Disable autofix for code correction to avoid recursion.
      correction = querying.query(
          CodeWithError(code=code, error=error), CorrectedCode, lm=lm, autofix=0
      )
    except pg.coding.CodeError:
      break

    code = correction.corrected_code
    result, error = result_and_error(code)
    if error is None:
      return (result, code) if returns_code else result

  raise pg.coding.CodeError(
      code,
      RuntimeError(
          f"Cannot correct code after {num_attempts} attempts. "
          f"Last error: {error}."
      ),
  )


def correct(
    code: str,
    error: str | None = None,
    *,
    global_vars: dict[str, Any] | None = None,
    lm: lf.LanguageModel = lf.contextual(),
    max_attempts: int = 5,
    sandbox: bool | None = None,
    timeout: int | None = 5,
) -> str:
  """Correct code with a language model via self-play.

  Args:
    code: The source code that may or may not be problematic.
    error: An optional initial error for `code` when it's problematic, usually
      caught from elsewhere when it ran. If None, code will be executed once to
      verify if its good and obtain a feedback error message.
    global_vars: A dict of str to value as the global variables that could be
      accessed within the corrected code.
    lm: Language model to be used. If not specified, it will try to use the `lm`
      under `lf.context`.
    max_attempts: Max number of attempts for the correction.
    sandbox: If True, run code in sandbox; If False, run code in current
      process. If None, run in sandbox first, if the output could not be
      serialized and pass to current process, run the code again in current
      process.
    timeout: The timeout for running the corrected code. If None, there is no
      timeout. Applicable only when sandbox is set to True.

  Returns:
    The final correct code if corrections are successful.

  Raises:
    `lf.CodeError`: If code cannot be corrected after `max_attempts`.
  """
  return run_with_correction(
      code,
      error=error,
      global_vars=global_vars,
      lm=lm,
      max_attempts=max_attempts,
      sandbox=sandbox,
      timeout=timeout,
      returns_code=True,
  )[1]


def _error_feedback_str(error: Exception) -> str:
  """Returns the error str for feedback."""
  if isinstance(error, pg.coding.CodeError):
    return pg.decolor(error.format(include_complete_code=False))
  else:
    return f"Encountered {error.__class__.__name__}: {error}"


def _maybe_custom_validate(result: Any) -> Any:
  """Apply custom validation through __validate_generation__ method."""
  if isinstance(result, dict) and "__result__" in result:
    r = result["__result__"]
  else:
    r = result

  if hasattr(r, "__validate__"):
    r.__validate__()
  return result



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for langfun.core.concurrent."""

import collections
from concurrent import futures
import contextlib
import io
import sys
import time
import unittest
from langfun.core import component
from langfun.core import concurrent
import pyglove as pg


class A(component.Component):
  x: int = 1
  y: int = component.contextual()


class RetryErrorTest(unittest.TestCase):

  def test_basics(self):
    def foo():
      pass

    retry_error = concurrent.RetryError(
        foo, [ValueError('abc'), ValueError('def'), ValueError('ghi')], [1, 2]
    )

    self.assertEqual(retry_error.attempts, 3)
    self.assertRegex(
        repr(retry_error),
        r'RetryError\(func=.*, errors=.*, wait_intervals=\[1, 2\]\)',
    )
    self.assertRegex(
        str(retry_error),
        (
            r'Calling .* failed after 3 attempts \(wait time: 1, 2 seconds\). '
            r'Last error: ghi'
        ),
    )

  def test_eq(self):
    f = lambda x: x
    e = ValueError()
    self.assertEqual(
        concurrent.RetryError(f, [e, e], [1]),
        concurrent.RetryError(f, [e, e], [1]),
    )
    self.assertNotEqual(
        concurrent.RetryError(f, [e, e], [1]),
        1,
    )
    self.assertNotEqual(
        concurrent.RetryError(f, [e, e], [1]),
        concurrent.RetryError(f, [e, e], [2]),
    )
    # Test hashing.
    self.assertEqual(
        hash(concurrent.RetryError(f, [e, e], [1])),
        hash(concurrent.RetryError(f, [e, e], [1])),
    )
    self.assertNotEqual(
        hash(concurrent.RetryError(f, [e, e], [1])),
        hash(concurrent.RetryError(f, [e, e], [2])),
    )


class RetryTest(unittest.TestCase):

  def assert_retry(self, func, expected_attempts, expected_wait_intervals):
    with pg.catch_errors(concurrent.RetryError) as error_context:
      func()

    self.assertIsNotNone(error_context.error)
    self.assertEqual(error_context.error.attempts, expected_attempts)
    self.assertEqual(
        error_context.error.wait_intervals, expected_wait_intervals
    )

  def test_retry_with_static_interval(self):
    def foo():
      raise ValueError('Intentional error.')

    foo_with_retry = concurrent.with_retry(
        foo,
        ValueError,
        max_attempts=3,
        retry_interval=1,
        exponential_backoff=False,
    )
    self.assert_retry(foo_with_retry, 3, [1, 1])

  def test_retry_with_interval_range(self):
    def foo():
      raise ValueError('Intentional error.')

    foo_with_retry = concurrent.with_retry(
        foo,
        ValueError,
        max_attempts=3,
        retry_interval=(1, 5),
        exponential_backoff=False,
        seed=1,
    )
    self.assert_retry(foo_with_retry, 3, [2, 5])

  def test_retry_with_exponential_backoff(self):
    def foo():
      raise ValueError('Intentional error.')

    foo_with_retry = concurrent.with_retry(
        foo, ValueError, max_attempts=4, retry_interval=1,
    )
    self.assert_retry(foo_with_retry, 4, [1, 2, 4])

  def test_retry_with_max_retry_interval(self):
    def foo():
      raise ValueError('Intentional error.')

    foo_with_retry = concurrent.with_retry(
        foo, ValueError, max_attempts=4, retry_interval=1, max_retry_interval=3,
    )
    self.assert_retry(foo_with_retry, 4, [1, 2, 3])

  def test_retry_with_uncaught_exception(self):
    def foo():
      raise ValueError('Intentional error.')

    foo_with_retry = concurrent.with_retry(
        foo, KeyError, max_attempts=4, retry_interval=1
    )

    with self.assertRaises(ValueError):
      foo_with_retry()

  def test_retry_with_job(self):
    count = 0

    def foo():
      nonlocal count
      count += 1
      if count < 3:
        raise ValueError('Foo temporary error.')
      return 'Success'

    job = concurrent.Job(
        foo,
        retry_on_errors=ValueError,
        retry_interval=1,
    )
    job()
    self.assertEqual(job.result, 'Success')
    self.assertEqual(
        [retry_entry.wait_interval for retry_entry in job.retry_entries],
        [0, 1, 2],
    )
    self.assertIsInstance(job.retry_entries[0].error, ValueError)
    self.assertIsInstance(job.retry_entries[1].error, ValueError)
    self.assertIsNone(job.retry_entries[2].error)


class ConcurrentExecuteTest(unittest.TestCase):

  def test_concurrent_execute(self):
    def fun(a):
      return a.x * a.y

    with component.context(y=2):
      self.assertEqual(concurrent.concurrent_execute(fun, [A(1), A(2)]), [2, 4])

  def test_concurrent_execute_with_a_single_worker(self):
    def fun(a):
      return a.x * a.y

    with component.context(y=2):
      self.assertEqual(
          concurrent.concurrent_execute(fun, [A(1), A(2)], max_workers=1),
          [2, 4],
      )

  def test_concurrent_execute_with_external_executor(self):
    def fun(a):
      return a.x * a.y

    executor = futures.ThreadPoolExecutor(max_workers=2)
    with component.context(y=2):
      self.assertEqual(
          concurrent.concurrent_execute(fun, [A(1), A(2)], executor=executor),
          [2, 4])

    # Making sure the executor could be reused.
    with component.context(y=2):
      self.assertEqual(
          concurrent.concurrent_execute(fun, [A(2), A(4)]), [4, 8])


class ProgressTest(unittest.TestCase):

  def test_progress(self):
    p = concurrent.Progress(total=10)
    self.assertEqual(p.total, 10)
    self.assertEqual(p.succeeded, 0)
    self.assertEqual(p.failed, 0)
    self.assertEqual(p.completed, 0)
    self.assertEqual(p.success_rate, 0)
    self.assertEqual(p.failure_rate, 0)
    self.assertEqual(p.avg_duration, 0)

    def fun(x):
      time.sleep(x)
      return x

    def fun2(unused_x):
      raise ValueError('Intentional error.')

    job1 = concurrent.Job(fun, (1,))
    job2 = concurrent.Job(fun2, (2,))
    job1()
    job2()

    p.update(job1)
    self.assertEqual(p.succeeded, 1)
    self.assertEqual(p.failed, 0)
    self.assertEqual(p.completed, 1)
    self.assertEqual(p.success_rate, 1)
    self.assertEqual(p.failure_rate, 0)
    self.assertGreater(p.avg_duration, 0.5)
    self.assertIs(p.job, job1)
    self.assertIsNone(p.last_error)

    p.update(job2)
    self.assertEqual(p.succeeded, 1)
    self.assertEqual(p.failed, 1)
    self.assertEqual(p.completed, 2)
    self.assertEqual(p.success_rate, 0.5)
    self.assertEqual(p.failure_rate, 0.5)
    self.assertIs(p.job, job2)
    self.assertIs(p.last_error, job2.error)


class ProgressControlTest(unittest.TestCase):

  def test_noop(self):
    concurrent.progress_bar = None
    ctrl = concurrent._progress_control(100, 'noop', 'blue', None)
    self.assertIsInstance(ctrl, concurrent._NoopProgressControl)
    string_io = io.StringIO()
    with contextlib.redirect_stderr(string_io):
      ctrl.update(1)
      ctrl.refresh()
    self.assertEqual(string_io.getvalue(), '')
    concurrent.progress_bar = 'tqdm'

  def test_console(self):
    concurrent.progress_bar = 'console'
    ctrl = concurrent._progress_control(100, 'foo', 'blue', None)
    self.assertIsInstance(ctrl, concurrent._ConsoleProgressControl)
    string_io = io.StringIO()
    with contextlib.redirect_stderr(string_io):
      ctrl.set_status('bar')
      ctrl.update(10)
      ctrl.refresh()
    self.assertEqual(
        string_io.getvalue(),
        '\x1b[1m\x1b[31mfoo\x1b[0m: \x1b[34m10% (10/100)\x1b[0m : bar\n'
    )
    concurrent.progress_bar = 'tqdm'

  def test_tqdm(self):
    concurrent.progress_bar = 'tqdm'
    string_io = io.StringIO()
    with contextlib.redirect_stderr(string_io):
      ctrl = concurrent._progress_control(100, 'foo', 'blue', None)
      self.assertIsInstance(ctrl, concurrent._TqdmProgressControl)
      ctrl.update(10)
      ctrl.refresh()
    self.assertIn('10/100', string_io.getvalue())

    tqdm = concurrent.tqdm
    concurrent.tqdm = None
    with self.assertRaisesRegex(RuntimeError, 'install package "tqdm"'):
      _ = concurrent._progress_control(100, 'foo', 'blue', None)
    concurrent.tqdm = tqdm

  def test_unsupported(self):
    concurrent.progress_bar = 'unknown'
    with self.assertRaisesRegex(ValueError, 'Unsupported progress bar type'):
      _ = concurrent._progress_control(100, 'foo', 'blue', None)
    concurrent.progress_bar = 'tqdm'


class ProgressBarTest(unittest.TestCase):

  def test_multithread_support(self):
    string_io = io.StringIO()
    with contextlib.redirect_stderr(string_io):
      bar_id = concurrent.ProgressBar.install(None, 5)
      def fun(x):
        del x
        concurrent.ProgressBar.update(bar_id, 1, status=None)

      for _ in concurrent.concurrent_execute(fun, range(5)):
        concurrent.ProgressBar.refresh()
      concurrent.ProgressBar.uninstall(bar_id)
    output_str = string_io.getvalue()
    self.assertIn('100%', output_str)
    self.assertIn('5/5', output_str)

  def test_report(self):
    string_io = io.StringIO()
    with contextlib.redirect_stderr(string_io):
      bar_id = concurrent.ProgressBar.install(None, 4)
      concurrent.ProgressBar.update(bar_id, 1, status=None)
      concurrent.ProgressBar.update(bar_id, 1, status='hello')
      concurrent.ProgressBar.update(bar_id, color='green')
      concurrent.ProgressBar.update(bar_id, 2, status=dict(x=1))
      with self.assertRaisesRegex(ValueError, 'Unsupported status'):
        concurrent.ProgressBar.update(bar_id, 0, status=1)
      concurrent.ProgressBar.uninstall(bar_id)
      sys.stderr.flush()
      time.sleep(1)
    self.assertIn('1/4', string_io.getvalue())
    # TODO(daiyip): Re-enable once flakiness is fixed.
    # self.assertIn('2/4', string_io.getvalue())
    # self.assertIn('hello', string_io.getvalue())
    # self.assertNotIn('3/4', string_io.getvalue())
    # self.assertIn('4/4', string_io.getvalue())
    # self.assertIn('x=1', string_io.getvalue())


class ConcurrentMapTest(unittest.TestCase):
  def test_concurrent_map_raise_on_error(self):
    error = ValueError()

    def fun(x):
      time.sleep(x)
      if x == 2:
        raise error
      return x**2

    with component.context(y=2):
      it = concurrent.concurrent_map(fun, [1, 2, 3], silence_on_errors=KeyError)
      self.assertEqual(next(it), (1, 1, None))
      with self.assertRaises(ValueError):
        _ = next(it)

    # Once error has raisen. The next call to `next` will raise StopIteration.
    with self.assertRaises(StopIteration):
      next(it)

  def test_concurrent_map_retry_on_error(self):
    error = ValueError()
    call_time = collections.defaultdict(int)

    def fun(x):
      call_time[x] += 1
      if call_time[x] >= x:
        return x ** 2
      raise error

    with component.context(y=2):
      self.assertEqual(
          set(
              concurrent.concurrent_map(
                  fun,
                  [1, 2, 3],
                  retry_on_errors=ValueError,
                  retry_interval=1,
                  silence_on_errors=concurrent.RetryError,
                  max_attempts=2,
              )
          ),
          set([
              (1, 1, None),
              (2, 4, None),
              (
                  3,
                  pg.MISSING_VALUE,
                  concurrent.RetryError(fun, [error, error], [1]),
              ),
          ]),
      )

  def test_concurrent_map_silence_on_errors(self):
    error = ValueError()

    def fun(x):
      if x == 2:
        raise error
      return x**2

    with component.context(y=2):
      self.assertEqual(
          set(
              concurrent.concurrent_map(
                  fun, [1, 2, 3], silence_on_errors=ValueError
              )
          ),
          set([
              (1, 1, None),
              (2, pg.MISSING_VALUE, error),
              (3, 9, None),
          ]),
      )

  def test_concurrent_map_with_async_complete(self):
    def fun(x):
      time.sleep(3 - x)
      return x

    with component.context(y=2):
      self.assertEqual(
          list(concurrent.concurrent_map(fun, [1, 2, 3])),
          [
              (3, 3, None),
              (2, 2, None),
              (1, 1, None),
          ],
      )

  def test_concurrent_map_with_ordering(self):
    def fun(x):
      time.sleep(3 - x)
      return x

    with component.context(y=2):
      self.assertEqual(
          list(concurrent.concurrent_map(fun, [1, 2, 3], ordered=True)),
          [
              (1, 1, None),
              (2, 2, None),
              (3, 3, None),
          ],
      )

  def test_concurrent_map_with_external_executor(self):
    def fun(x):
      return x

    executor = futures.ThreadPoolExecutor(max_workers=2)
    self.assertEqual(
        list(concurrent.concurrent_map(
            fun, [1, 2, 3], executor=executor, ordered=True)),
        [
            (1, 1, None),
            (2, 2, None),
            (3, 3, None),
        ],
    )
    self.assertEqual(
        list(concurrent.concurrent_map(
            fun, [4, 5, 6], executor=executor, ordered=True)),
        [
            (4, 4, None),
            (5, 5, None),
            (6, 6, None),
        ],
    )

  def test_concurrent_map_with_order_and_raise_on_errors(self):
    error = ValueError()

    def fun(x):
      if x == 2:
        raise error
      return x**2

    with component.context(y=2):
      it = concurrent.concurrent_map(
          fun, [1, 2, 3], ordered=True, silence_on_errors=KeyError)
      self.assertEqual(next(it)[1], 1)

      with self.assertRaises(ValueError):
        _ = next(it)

  def test_concurrent_map_with_order_and_timeout(self):
    def fun(x):
      time.sleep(3 - x)
      return x

    self.assertEqual(
        [
            (i, o)
            for i, o, _ in concurrent.concurrent_map(
                fun, [-1, 2, 3], ordered=True, timeout=1.5
            )
        ],
        [
            (-1, pg.MISSING_VALUE),
            (2, 2),
            (3, 3),
        ],
    )

  def test_concurent_map_unordered_with_timeout(self):
    def fun(x):
      time.sleep(x)
      return x

    self.assertEqual(
        [
            (i, o)
            for i, o, _ in concurrent.concurrent_map(
                fun, [5, 2, 1, 4], timeout=3
            )
        ],
        [
            (1, 1),
            (2, 2),
            (5, pg.MISSING_VALUE),
            (4, pg.MISSING_VALUE),
        ],
    )
    with self.assertRaises(TimeoutError):
      next(concurrent.concurrent_map(
          fun, [5, 3], timeout=1, silence_on_errors=None))

  def test_concurent_map_unordered_with_timeout_less_worker(self):
    def fun(x):
      time.sleep(x)
      return x

    self.assertEqual(
        [
            (i, o)
            for i, o, _ in concurrent.concurrent_map(
                fun, [5, 2, 1, 6], timeout=3, max_workers=1
            )
        ],
        [
            (5, pg.MISSING_VALUE),
            (2, 2),
            (1, 1),
            (6, pg.MISSING_VALUE),
        ],
    )

  def test_concurrent_map_with_showing_progress(self):
    def fun(x):
      with pg.timeit('foo'):
        if x == 2:
          raise ValueError('Intentional error.')
        time.sleep(x)
        return x

    string_io = io.StringIO()
    with contextlib.redirect_stderr(string_io):
      output = sorted([
          (i, o) for i, o, _ in concurrent.concurrent_map(
              fun, [1, 2, 3], timeout=1.5, max_workers=1, show_progress=True
          )
      ], key=lambda x: x[0])
      string_io.flush()
    self.assertEqual(   # pylint: disable=g-generic-assert
        output,
        [
            (1, 1),
            (2, pg.MISSING_VALUE),
            (3, pg.MISSING_VALUE),
        ],
    )
    output = string_io.getvalue()
    self.assertIn('100%', output)

  def test_concurrent_map_with_showing_progress_and_status_fn(self):
    def fun(x):
      if x == 2:
        raise ValueError('Intentional error.')
      time.sleep(x)
      return x

    bar_id = concurrent.ProgressBar.install(None, 3)
    string_io = io.StringIO()
    with contextlib.redirect_stderr(string_io):
      output = sorted([
          (i, o) for i, o, _ in concurrent.concurrent_map(
              fun, [1, 2, 3], timeout=1.5, max_workers=1,
              show_progress=bar_id, status_fn=lambda p: dict(x=1, y=1)
          )
      ], key=lambda x: x[0])

    self.assertEqual(  # pylint: disable=g-generic-assert
        output,
        [
            (1, 1),
            (2, pg.MISSING_VALUE),
            (3, pg.MISSING_VALUE),
        ],
    )
    concurrent.ProgressBar.uninstall(bar_id)
    self.assertIn('100%', string_io.getvalue())


class ExecutorPoolTest(unittest.TestCase):

  def test_pool(self):
    pool = concurrent.ExecutorPool()
    executor1 = futures.ThreadPoolExecutor()
    self.assertIs(pool.executor_from(executor1), executor1)

    executor2 = pool.executor_from('executor2', max_workers=1)
    self.assertIsInstance(executor2, futures.ThreadPoolExecutor)
    self.assertIs(pool.get('executor2'), executor2)
    self.assertEqual(pool.resource_ids, ['executor2'])

    executor3 = pool.executor_from(None, max_workers=1)
    self.assertIsInstance(executor3, futures.ThreadPoolExecutor)
    self.assertEqual(pool.resource_ids, ['executor2'])

    with self.assertRaises(ValueError):
      pool.executor_from(1)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for langfun.core.structured.completion."""

import asyncio
import inspect
import unittest

import langfun.core as lf
from langfun.core import modalities
from langfun.core.llms import fake
from langfun.core.structured import completion
from langfun.core.structured import mapping
from langfun.core.structured import schema as schema_lib
import pyglove as pg


class Activity(pg.Object):
  description: str


class Itinerary(pg.Object):
  day: pg.typing.Int[1, None]
  type: pg.typing.Annotated[
      pg.typing.Enum['daytime', 'nighttime'],
      'Type of itinerary.'
  ]
  activities: list[Activity]


class TripPlan(pg.Object):
  place: str
  itineraries: list[Itinerary]


class CompleteStructureTest(unittest.TestCase):

  def test_render_no_examples(self):
    l = completion._CompleteStructure()
    input_value = schema_lib.mark_missing(
        TripPlan.partial(
            place='San Francisco',
            itineraries=[
                Itinerary.partial(day=1),
                Itinerary.partial(day=2),
                Itinerary.partial(day=3),
            ],
        )
    )
    self.assertEqual(
        l.render(input=input_value).text,
        inspect.cleandoc("""
            Please generate the OUTPUT_OBJECT by completing the MISSING fields from the last INPUT_OBJECT.

            INSTRUCTIONS:
            1. Each MISSING field contains a Python annotation, please fill the value based on the annotation.
            2. Classes for the MISSING fields are defined under CLASS_DEFINITIONS.

            INPUT_OBJECT:
              ```python
              Answer(
                question='1 + 1 =',
                answer=MISSING(int)
              )
              ```

            OUTPUT_OBJECT:
              ```python
              Answer(
                question='1 + 1 =',
                answer=2
              )
              ```

            INPUT_OBJECT:
              ```python
              TripPlan(
                place='San Francisco',
                itineraries=[
                  Itinerary(
                    day=1,
                    # Type of itinerary.
                    type=MISSING(Literal['daytime', 'nighttime']),
                    activities=MISSING(list[Activity])
                  ),
                  Itinerary(
                    day=2,
                    # Type of itinerary.
                    type=MISSING(Literal['daytime', 'nighttime']),
                    activities=MISSING(list[Activity])
                  ),
                  Itinerary(
                    day=3,
                    # Type of itinerary.
                    type=MISSING(Literal['daytime', 'nighttime']),
                    activities=MISSING(list[Activity])
                  )
                ]
              )
              ```

            CLASS_DEFINITIONS:
              ```python
              class Activity:
                description: str
              ```

            OUTPUT_OBJECT:
            """),
    )

  def test_render_no_class_definitions(self):
    l = completion._CompleteStructure()
    input_value = schema_lib.mark_missing(
        TripPlan.partial(
            place='San Francisco',
            itineraries=[
                Itinerary.partial(day=1, activities=[Activity.partial()]),
                Itinerary.partial(day=2, activities=[Activity.partial()]),
                Itinerary.partial(day=3, activities=[Activity.partial()]),
            ],
        )
    )
    self.assertEqual(
        l.render(input=input_value).text,
        inspect.cleandoc("""
            Please generate the OUTPUT_OBJECT by completing the MISSING fields from the last INPUT_OBJECT.

            INSTRUCTIONS:
            1. Each MISSING field contains a Python annotation, please fill the value based on the annotation.
            2. Classes for the MISSING fields are defined under CLASS_DEFINITIONS.

            INPUT_OBJECT:
              ```python
              Answer(
                question='1 + 1 =',
                answer=MISSING(int)
              )
              ```

            OUTPUT_OBJECT:
              ```python
              Answer(
                question='1 + 1 =',
                answer=2
              )
              ```

            INPUT_OBJECT:
              ```python
              TripPlan(
                place='San Francisco',
                itineraries=[
                  Itinerary(
                    day=1,
                    # Type of itinerary.
                    type=MISSING(Literal['daytime', 'nighttime']),
                    activities=[
                      Activity(
                        description=MISSING(str)
                      )
                    ]
                  ),
                  Itinerary(
                    day=2,
                    # Type of itinerary.
                    type=MISSING(Literal['daytime', 'nighttime']),
                    activities=[
                      Activity(
                        description=MISSING(str)
                      )
                    ]
                  ),
                  Itinerary(
                    day=3,
                    # Type of itinerary.
                    type=MISSING(Literal['daytime', 'nighttime']),
                    activities=[
                      Activity(
                        description=MISSING(str)
                      )
                    ]
                  )
                ]
              )
              ```

            OUTPUT_OBJECT:
            """),
    )

  def test_render_with_examples(self):
    l = completion._CompleteStructure()
    input_value = schema_lib.mark_missing(
        TripPlan.partial(
            place='San Francisco',
            itineraries=[
                Itinerary.partial(day=1),
                Itinerary.partial(day=2),
                Itinerary.partial(day=3),
            ],
        )
    )
    self.assertEqual(
        l.render(input=input_value).text,
        inspect.cleandoc("""
          Please generate the OUTPUT_OBJECT by completing the MISSING fields from the last INPUT_OBJECT.

          INSTRUCTIONS:
          1. Each MISSING field contains a Python annotation, please fill the value based on the annotation.
          2. Classes for the MISSING fields are defined under CLASS_DEFINITIONS.

          INPUT_OBJECT:
            ```python
            Answer(
              question='1 + 1 =',
              answer=MISSING(int)
            )
            ```

          OUTPUT_OBJECT:
            ```python
            Answer(
              question='1 + 1 =',
              answer=2
            )
            ```

          INPUT_OBJECT:
            ```python
            TripPlan(
              place='San Francisco',
              itineraries=[
                Itinerary(
                  day=1,
                  # Type of itinerary.
                  type=MISSING(Literal['daytime', 'nighttime']),
                  activities=MISSING(list[Activity])
                ),
                Itinerary(
                  day=2,
                  # Type of itinerary.
                  type=MISSING(Literal['daytime', 'nighttime']),
                  activities=MISSING(list[Activity])
                ),
                Itinerary(
                  day=3,
                  # Type of itinerary.
                  type=MISSING(Literal['daytime', 'nighttime']),
                  activities=MISSING(list[Activity])
                )
              ]
            )
            ```

          CLASS_DEFINITIONS:
            ```python
            class Activity:
              description: str
            ```

          OUTPUT_OBJECT:"""),
    )

  def test_invocation(self):
    structured_response = inspect.cleandoc("""
        ```python
        TripPlan(
          place='San Francisco',
          itineraries=[
            Itinerary(
                day=1,
                type='daytime',
                activities=[
                    Activity(description='Arrive in San Francisco and check into your hotel.'),
                    Activity(description='Take a walk around Fisherman\\'s Wharf and have dinner at one of the many seafood restaurants.'),
                    Activity(description='Visit Pier 39 and see the sea lions.'),
                ], 
                ),
            Itinerary(
                day=2,
                type='daytime',
                activities=[
                    Activity(description='Take a ferry to Alcatraz Island and tour the infamous prison.'),
                    Activity(description='Take a walk across the Golden Gate Bridge.'),
                    Activity(description='Visit the Japanese Tea Garden in Golden Gate Park.'),
                ], 
                ),
            Itinerary(
                day=3,
                type='daytime',
                activities=[
                    Activity(description='Visit the de Young Museum and see the collection of American art.'),
                    Activity(description='Visit the San Francisco Museum of Modern Art.'),
                    Activity(description='Take a cable car ride.'),
                ], 
                ),
          ]
        )
        ```
        """)

    with lf.context(
        lm=fake.StaticSequence(
            [structured_response],
        ),
        override_attrs=True,
    ):
      r = completion.complete(
          TripPlan.partial(
              place='San Francisco',
              itineraries=[
                  Itinerary.partial(day=1),
                  Itinerary.partial(day=2),
                  Itinerary.partial(day=3),
              ],
          )
      )
      # pylint: disable=line-too-long
      self.assertEqual(
          r,
          TripPlan(
              place='San Francisco',
              itineraries=[
                  Itinerary(
                      day=1,
                      type='daytime',
                      activities=[
                          Activity(
                              description=(
                                  'Arrive in San Francisco and check into your'
                                  ' hotel.'
                              )
                          ),
                          Activity(
                              description=(
                                  "Take a walk around Fisherman's Wharf and"
                                  ' have dinner at one of the many seafood'
                                  ' restaurants.'
                              )
                          ),
                          Activity(
                              description='Visit Pier 39 and see the sea lions.'
                          ),
                      ],
                  ),
                  Itinerary(
                      day=2,
                      type='daytime',
                      activities=[
                          Activity(
                              description=(
                                  'Take a ferry to Alcatraz Island and tour the'
                                  ' infamous prison.'
                              )
                          ),
                          Activity(
                              description=(
                                  'Take a walk across the Golden Gate Bridge.'
                              )
                          ),
                          Activity(
                              description=(
                                  'Visit the Japanese Tea Garden in Golden Gate'
                                  ' Park.'
                              )
                          ),
                      ],
                  ),
                  Itinerary(
                      day=3,
                      type='daytime',
                      activities=[
                          Activity(
                              description=(
                                  'Visit the de Young Museum and see the'
                                  ' collection of American art.'
                              )
                          ),
                          Activity(
                              description=(
                                  'Visit the San Francisco Museum of Modern'
                                  ' Art.'
                              )
                          ),
                          Activity(description='Take a cable car ride.'),
                      ],
                  ),
              ],
          ),
      )
      # pylint: enable=line-too-long

  def test_invocation_with_modality(self):
    class Animal(pg.Object):
      image: modalities.Image
      name: str

    input_value = schema_lib.mark_missing(
        Animal.partial(
            modalities.Image.from_bytes(b'image_of_elephant'),
        )
    )
    l = completion._CompleteStructure(
        input=input_value,
        examples=[
            mapping.MappingExample(
                input=Animal.partial(
                    modalities.Image.from_bytes(b'image_of_rabbit')
                ),
                output=Animal(
                    modalities.Image.from_bytes(b'image_of_rabbit'),
                    'rabbit',
                ),
            )
        ],
    )
    lm_input = l.render()
    self.maxDiff = None
    self.assertEqual(
        lm_input.text,
        inspect.cleandoc("""
            Please generate the OUTPUT_OBJECT by completing the MISSING fields from the last INPUT_OBJECT.

            INSTRUCTIONS:
            1. Each MISSING field contains a Python annotation, please fill the value based on the annotation.
            2. Classes for the MISSING fields are defined under CLASS_DEFINITIONS.

            INPUT_OBJECT:
              ```python
              Answer(
                question='1 + 1 =',
                answer=MISSING(int)
              )
              ```

            OUTPUT_OBJECT:
              ```python
              Answer(
                question='1 + 1 =',
                answer=2
              )
              ```

            INPUT_OBJECT:
              ```python
              Animal(
                image=ModalityRef(
                  name='examples[0].input.image'
                ),
                name=MISSING(str)
              )
              ```

            MODALITY_REFERENCES:
              {
                'examples[0].input.image': <<[[examples[0].input.image]]>>
              }

            OUTPUT_OBJECT:
              ```python
              Animal(
                image=ModalityRef(
                  name='examples[0].output.image'
                ),
                name='rabbit'
              )
              ```


            INPUT_OBJECT:
              ```python
              Animal(
                image=ModalityRef(
                  name='input.image'
                ),
                name=MISSING(str)
              )
              ```

            MODALITY_REFERENCES:
              {
                'input.image': <<[[input.image]]>>
              }

            OUTPUT_OBJECT:
            """),
    )
    self.assertTrue(
        pg.eq(
            {
                'examples': lm_input.get('examples'),
                'input': lm_input.get('input'),
            },
            {
                'examples': [
                    mapping.MappingExample(
                        input=Animal.partial(
                            image=modalities.Image.from_bytes(
                                b'image_of_rabbit'
                            )
                        ),
                        output=Animal.partial(
                            image=modalities.Image.from_bytes(
                                b'image_of_rabbit'
                            ),
                            name='rabbit',
                        ),
                    )
                ],
                'input': Animal(
                    image=modalities.Image.from_bytes(b'image_of_elephant'),
                    name=schema_lib.MISSING,
                ),
            },
        )
    )
    lm_output = l(
        input=input_value,
        lm=fake.StaticResponse(inspect.cleandoc("""
            ```python
            Animal(
              image=ModalityRef(
                name='input.image'
              ),
              name='elephant'
            )
            ```
            """)),
    )
    self.assertTrue(
        pg.eq(
            lm_output.result,
            Animal(
                image=modalities.Image.from_bytes(b'image_of_elephant'),
                name='elephant',
            ),
        )
    )

  def test_autofix(self):
    class Solution(pg.Object):
      question: str
      answer: int

    lm = fake.StaticSequence(
        [
            "Solution(question='Compute 1 + 1', answer=2",
            inspect.cleandoc("""
            CorrectedCode(
                corrected_code='Solution(question=\\\'Compute 1 + 1\\\', answer=2)',
            )
            """),
        ],
        debug=True,
    )
    self.assertEqual(
        completion.complete(
            Solution.partial('Compute 1 + 1'), lm=lm, autofix=3
        ),
        Solution(question='Compute 1 + 1', answer=2),
    )

  def test_returns_message(self):
    self.assertEqual(
        completion.complete(
            Activity.partial(),
            lm=fake.StaticSequence(['Activity(description="foo")']),
            returns_message=True),
        lf.AIMessage(
            text='Activity(description="foo")',
            result=Activity(description='foo'),
            score=1.0,
            is_cached=False,
            logprobs=None,
            usage=lf.LMSamplingUsage(553, 27, 580),
            tags=['lm-response', 'lm-output', 'transformed']
        )
    )

  def test_using_the_same_lm_instance(self):
    lm = fake.StaticSequence(['Activity(description="foo")'])
    self.assertEqual(
        completion.complete(Activity.partial(), lm=lm), Activity('foo')
    )
    with self.assertRaises(IndexError):
      completion.complete(Activity.partial(), lm=lm)

  def test_bad_call(self):
    with self.assertRaisesRegex(
        ValueError, '.*must contain a least .* missing'
    ):
      completion.complete(Activity('foo'), lm=fake.StaticResponse(''))

  def test_bad_transform(self):
    with lf.context(
        lm=fake.StaticSequence(['Activity(description=1)']),
        override_attrs=True,
    ):
      with self.assertRaisesRegex(
          mapping.MappingError,
          'Expect .* but encountered .*',
      ):
        completion.complete(Activity.partial(), autofix=0)

  def test_default(self):
    with lf.context(
        lm=fake.StaticSequence(['Activity(description=1)']),
        override_attrs=True,
    ):
      self.assertIsNone(completion.complete(Activity.partial(), None))

  def test_acomplete(self):
    response = """
    ```python
        TripPlan(
          place='San Francisco',
          itineraries=[
            Itinerary(
                day=1,
                type='daytime',
                activities=[
                    Activity(description='Arrive in San Francisco and check into your hotel.'),
                    Activity(description='Take a walk around Fisherman\\'s Wharf and have dinner at one of the many seafood restaurants.'),
                    Activity(description='Visit Pier 39 and see the sea lions.'),
                ], 
            ),
          ]
        )
    ```
    """
    with lf.context(
        lm=fake.StaticSequence(
            [response],
        ),
        override_attrs=True,
    ):
      r = completion.acomplete(
          TripPlan.partial(
              place='San Francisco',
              itineraries=[
                  Itinerary.partial(day=1),
              ],
          )
      )
      self.assertIsInstance(asyncio.run(r), TripPlan)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for demonstration."""

import inspect
import unittest
from langfun.core.templates.completion import Completion
from langfun.core.templates.demonstration import Demonstration


class DemonstrationTest(unittest.TestCase):

  def test_bool(self):
    self.assertTrue(Demonstration([
        Completion(prompt='hi', response='hello')
    ]))
    self.assertFalse(Demonstration(examples=[]))

  def test_examples_only(self):
    self.assertEqual(
        Demonstration([
            Completion(prompt='hi', lm_response='hello'),
            Completion(prompt='goodbye', lm_response='bye'),
        ]).render(),
        inspect.cleandoc("""
            Here are the examples for demonstrating this:

              Example 1:
                hi
                hello

              Example 2:
                goodbye
                bye
            """),
    )

  def test_examples_with_description(self):
    self.assertEqual(
        Demonstration(
            [
                Completion(prompt='hi', lm_response='hello'),
                Completion(prompt='goodbye', lm_response='bye'),
            ],
            'Here are a few examples to greet people:',
        ).render(),
        inspect.cleandoc("""
            Here are a few examples to greet people:

              Example 1:
                hi
                hello

              Example 2:
                goodbye
                bye
            """),
    )

  def test_call(self):
    with self.assertRaises(ValueError):
      _ = Demonstration(examples=[])()


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
import unittest

from langfun.core.eval.v2 import metric_values
import pyglove as pg


class RateTest(unittest.TestCase):

  def test_basic(self):
    rate = metric_values.Rate()
    self.assertEqual(rate.total, 0)
    self.assertTrue(math.isnan(float(rate)))
    self.assertEqual(pg.format(rate), 'n/a')
    rate.increment_total()
    self.assertEqual(rate.total, 1)
    self.assertEqual(float(rate), 0.0)
    rate.add(1, 1.0, 1.0)
    self.assertEqual(float(rate), 1.0)
    self.assertEqual(pg.format(rate, verbose=False), '100.0%')
    self.assertEqual(pg.format(rate, verbose=True), '100.0% (1/1)')
    self.assertEqual(
        rate.data_points, [metric_values.MetricValue.DataPoint(1, 1.0, 1.0)]
    )
    self.assertEqual(rate, 1.0)
    self.assertGreater(rate, 0.5)
    self.assertLess(rate, 1.5)
    self.assertEqual(
        rate,
        metric_values.Rate(
            [metric_values.MetricValue.DataPoint(1, 1.0, 1.0)], 1
        )
    )
    self.assertGreater(rate, metric_values.Rate([], 1))
    self.assertLess(metric_values.Rate([], 1), rate)

    rate.reset()
    self.assertEqual(rate.total, 0)
    self.assertTrue(math.isnan(float(rate)))


class AverageTest(unittest.TestCase):

  def test_basic(self):
    average = metric_values.Average()
    self.assertEqual(average.total, 0)
    self.assertTrue(math.isnan(float(average)))
    self.assertEqual(pg.format(average, verbose=False), 'n/a')
    average.add(1, 1.0, 0.5, increment_total=True)
    average.add(1, 0.0, 1.0, increment_total=True)
    self.assertEqual(average.total, 2)
    self.assertEqual(float(average), 0.25)
    self.assertEqual(pg.format(average, verbose=False), '0.250')
    self.assertEqual(pg.format(average, verbose=True), '0.250 (2/2)')
    self.assertEqual(
        average.data_points,
        [
            metric_values.MetricValue.DataPoint(1, 1.0, 0.5),
            metric_values.MetricValue.DataPoint(1, 0.0, 1.0),
        ]
    )
    average.reset()
    self.assertEqual(average.total, 0)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python code execution."""

from typing import Any

from langfun.core.coding.python import parsing
import pyglove as pg


context = pg.coding.context
CodeError = pg.coding.CodeError
CodePermission = pg.coding.CodePermission
permission = pg.coding.permission


def evaluate(
    code: str,
    *,
    global_vars: dict[str, Any] | None = None,
    permission: CodePermission | None = None,   # pylint: disable=redefined-outer-name
    returns_stdout: bool = False,
    outputs_intermediate: bool = False,
) -> Any | dict[str, Any]:
  """Executes Python code.

  Features:
    * Fine-grained execution policy for limiting what APIs could be executed.
      This eliminates the need for sandboxing.
    * It exposes both the final results and intermediate results (variables).

  Args:
    code: Python code to run.
    global_vars: An optional dict as the globals that could be referenced by the
      code.
    permission: Permission for the Python code to run.
    returns_stdout: If True, the stdout (a str) will be returned.
    outputs_intermediate: Applicable when returns_stdout is False. If True,
      intermediate output will be outputted as a dict, with the last line's
      value accessible by key '__result__' and the std output accessible by
      key '__stdout__'. Otherwise the value of the last line will be returned.

  Returns:
    The value of the last line of the code block. Or a dict of variable
    names of all locals to their evaluated values as the output of the code to
    run. The value for the last line can be accessed by key '__result__'. Or the
    stdout as a str.
  """
  return pg.coding.evaluate(
      parsing.clean(code),
      global_vars=global_vars,
      permission=permission,
      returns_stdout=returns_stdout,
      outputs_intermediate=outputs_intermediate,
  )


def run(
    code: str,
    *,
    global_vars: dict[str, Any] | None = None,
    permission: CodePermission | None = None,  # pylint: disable=redefined-outer-name
    returns_stdout: bool = False,
    outputs_intermediate: bool = False,
    sandbox: bool | None = None,
    timeout: float | None = None,
) -> Any | dict[str, Any]:
  """Executes Python code.

  Features:
    * Fine-grained execution policy for limiting what APIs could be executed.
      This eliminates the need for sandboxing.
    * It exposes both the final results and intermediate results (variables).

  Args:
    code: Python code to run.
    global_vars: An optional dict of
    permission: Permission for the Python code to run.
    returns_stdout: If True, the stdout (a str) will be returned.
    outputs_intermediate: Applicable when returns_stdout is False. If True,
      intermediate output will be outputted as a dict, with the last line's
      value accessible by key '__result__' and the std output accessible by
      key '__stdout__'. Otherwise the value of the last line will be returned.
    sandbox: If True, run code in sandbox; If False, run code in current
      process. If None, run in sandbox first, if the output could not be
      serialized and pass to current process, run the code again in current
      process.
    timeout: Execution timeout in seconds. If None, wait the code the complete.

  Returns:
    The value of the last line of the code block. Or a dict of variable
    names of all locals to their evaluated values as the output of the code to
    run. The value for the last line can be accessed by key '__result__'. Or the
    stdout as a str.

  Raises:
    TimeoutError: If the execution time exceeds the timeout.
    Exception: Exception  that are raised from the code.
  """
  return pg.coding.run(
      parsing.clean(code),
      global_vars=global_vars,
      permission=permission,
      returns_stdout=returns_stdout,
      outputs_intermediate=outputs_intermediate,
      sandbox=sandbox,
      timeout=timeout,
  )



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import tempfile
import threading
import time
from typing import Any
import unittest

from langfun.core.eval.v2 import eval_test_helper
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
from langfun.core.eval.v2 import runners as runners_lib  # pylint: disable=unused-import

import pyglove as pg


Runner = experiment_lib.Runner
Example = example_lib.Example
Experiment = experiment_lib.Experiment
Suite = experiment_lib.Suite
Plugin = experiment_lib.Plugin


class TestPlugin(Plugin):
  started_experiments: list[Experiment] = []
  completed_experiments: list[Experiment] = []
  skipped_experiments: list[Experiment] = []
  started_example_ids: list[int] = []
  completed_example_ids: list[int] = []
  skipped_example_ids: list[int] = []
  start_time: float | None = None
  complete_time: float | None = None

  def _on_bound(self):
    super()._on_bound()
    self._lock = threading.Lock()

  def on_run_start(self, runner: Runner, root: Experiment):
    del root
    with pg.notify_on_change(False), pg.allow_writable_accessors(True):
      self.start_time = time.time()

  def on_run_complete(self, runner: Runner, root: Experiment):
    del root
    with pg.notify_on_change(False), pg.allow_writable_accessors(True):
      self.complete_time = time.time()

  def on_experiment_start(self, runner: Runner, experiment: Experiment):
    del runner
    with pg.notify_on_change(False), self._lock:
      self.started_experiments.append(pg.Ref(experiment))

  def on_experiment_skipped(self, runner: Runner, experiment: Experiment):
    del runner
    with pg.notify_on_change(False), self._lock:
      self.skipped_experiments.append(pg.Ref(experiment))

  def on_experiment_complete(self, runner: Runner, experiment: Experiment):
    del runner
    with pg.notify_on_change(False), self._lock:
      self.completed_experiments.append(pg.Ref(experiment))

  def on_example_start(
      self, runner: Runner, experiment: Experiment, example: Example):
    del runner, experiment
    with pg.notify_on_change(False), self._lock:
      self.started_example_ids.append(example.id)

  def on_example_skipped(
      self, runner: Runner, experiment: Experiment, example: Example):
    del runner, experiment
    with pg.notify_on_change(False), self._lock:
      self.skipped_example_ids.append(example.id)

  def on_example_complete(
      self, runner: Runner, experiment: Experiment, example: Example):
    del runner, experiment
    with pg.notify_on_change(False), self._lock:
      self.completed_example_ids.append(example.id)


class RunnerTest(unittest.TestCase):

  def assert_same_list(self, actual: list[Any], expected: list[Any]):
    self.assertEqual(len(actual), len(expected))
    for i, (x, y) in enumerate(zip(actual, expected)):
      if x is not y:
        print(i, pg.diff(x, y))
      self.assertIs(x, y)

  def test_basic(self):
    plugin = TestPlugin()
    exp = eval_test_helper.test_experiment()
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_sequential_runner')
    run = exp.run(root_dir, runner='sequential', plugins=[plugin])

    self.assertIsNotNone(plugin.start_time)
    self.assertIsNotNone(plugin.complete_time)
    self.assertGreater(plugin.complete_time, plugin.start_time)

    self.assert_same_list(
        plugin.started_experiments,
        exp.nonleaf_nodes + exp.leaf_nodes
    )
    self.assert_same_list(
        plugin.completed_experiments,
        exp.leaf_nodes + list(reversed(exp.nonleaf_nodes))
    )
    self.assert_same_list(
        plugin.started_example_ids, list(range(1, 11)) * 6
    )
    self.assert_same_list(
        plugin.completed_example_ids, list(range(1, 11)) * 6
    )
    self.assert_same_list(plugin.skipped_experiments, [])
    self.assert_same_list(plugin.skipped_example_ids, [])
    self.assertTrue(
        pg.io.path_exists(os.path.join(run.output_root, 'run.json'))
    )

    for node in exp.nodes:
      self.assertTrue(node.progress.is_started)
      self.assertTrue(node.progress.is_completed)
      if node.is_leaf:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_completed, 10)
        self.assertEqual(node.progress.num_failed, 1)
      else:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_failed, 0)
        self.assertEqual(node.progress.num_processed, node.progress.num_total)

  def test_raise_if_has_error(self):
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_raise_if_has_error')
    exp = eval_test_helper.TestEvaluation()
    with self.assertRaisesRegex(ValueError, 'x should not be 5'):
      exp.run(
          root_dir, runner='sequential', plugins=[], raise_if_has_error=True
      )

    with self.assertRaisesRegex(ValueError, 'x should not be 5'):
      exp.run(root_dir, runner='parallel', plugins=[], raise_if_has_error=True)

  def test_example_ids(self):
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_example_ids')
    exp = eval_test_helper.test_experiment()
    plugin = TestPlugin()
    _ = exp.run(
        root_dir, runner='sequential', plugins=[plugin], example_ids=[5, 7, 9]
    )
    self.assertEqual(plugin.started_example_ids, [5, 7, 9] * 6)
    self.assertEqual(plugin.completed_example_ids, [5, 7, 9] * 6)

  def test_shuffle_inputs(self):
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_shuffle_inputs')
    exp = eval_test_helper.test_experiment()
    plugin = TestPlugin()
    run = exp.run(
        root_dir, runner='sequential', plugins=[plugin], shuffle_inputs=True
    )
    self.assertTrue(run.shuffle_inputs)

  def test_filter(self):
    plugin = TestPlugin()
    exp = eval_test_helper.test_experiment()
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_filter')

    _ = exp.run(
        root_dir, runner='sequential', plugins=[plugin],
        filter=lambda e: e.lm.offset != 0
    )
    self.assert_same_list(
        plugin.started_experiments,
        exp.nonleaf_nodes + exp.leaf_nodes[2:]
    )
    self.assert_same_list(
        plugin.skipped_experiments, exp.leaf_nodes[:2]
    )
    self.assert_same_list(
        plugin.completed_experiments,
        exp.leaf_nodes[2:] + [exp.children[1], exp]
    )

  def test_use_cache(self):
    @pg.functor()
    def test_inputs(num_examples: int = 10):
      return [
          pg.Dict(
              x=i // 2, y=(i // 2) ** 2,
              groundtruth=(i // 2 + (i // 2) ** 2)
          ) for i in range(num_examples)
      ]

    exp = eval_test_helper.TestEvaluation(
        inputs=test_inputs(num_examples=pg.oneof([2, 4]))
    )
    # Global cache.
    root_dir = os.path.join(tempfile.mkdtemp(), 'global_cache')
    run = exp.run(
        root_dir, 'new', runner='sequential', use_cache='global', plugins=[]
    )
    self.assertTrue(pg.io.path_exists(run.output_path_for(exp, 'cache.json')))
    self.assertEqual(exp.usage_summary.cached.total.num_requests, 4)
    self.assertEqual(exp.usage_summary.uncached.total.num_requests, 2)

    # Per-dataset cache.
    root_dir = os.path.join(tempfile.mkdtemp(), 'per_dataset')
    run = exp.run(
        root_dir, 'new', runner='sequential',
        use_cache='per_dataset', plugins=[]
    )
    for leaf in exp.leaf_nodes:
      self.assertTrue(
          pg.io.path_exists(run.output_path_for(leaf, 'cache.json'))
      )
    self.assertEqual(exp.usage_summary.cached.total.num_requests, 3)
    self.assertEqual(exp.usage_summary.uncached.total.num_requests, 3)

    # No cache.
    root_dir = os.path.join(tempfile.mkdtemp(), 'no')
    run = exp.run(root_dir, runner='sequential', use_cache='no', plugins=[])
    self.assertFalse(pg.io.path_exists(run.output_path_for(exp, 'cache.json')))
    for leaf in exp.leaf_nodes:
      self.assertFalse(
          pg.io.path_exists(run.output_path_for(leaf, 'cache.json'))
      )
    self.assertEqual(exp.usage_summary.cached.total.num_requests, 0)
    self.assertEqual(exp.usage_summary.uncached.total.num_requests, 6)


class ParallelRunnerTest(RunnerTest):

  def test_parallel_runner(self):
    plugin = TestPlugin()
    exp = eval_test_helper.test_experiment()
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_parallel_runner')
    run = exp.run(root_dir, runner='parallel', plugins=[plugin])

    self.assertIsNotNone(plugin.start_time)
    self.assertIsNotNone(plugin.complete_time)
    self.assertGreater(plugin.complete_time, plugin.start_time)

    self.assertEqual(
        len(plugin.started_experiments), len(exp.nodes)
    )
    self.assertEqual(
        len(plugin.completed_experiments), len(exp.nodes)
    )
    self.assertEqual(
        len(plugin.started_example_ids), 6 * 10
    )
    self.assertEqual(
        len(plugin.completed_example_ids), 6 * 10
    )
    self.assert_same_list(plugin.skipped_experiments, [])
    self.assert_same_list(plugin.skipped_example_ids, [])
    self.assertTrue(
        pg.io.path_exists(os.path.join(run.output_root, 'run.json'))
    )

    for node in exp.nodes:
      self.assertTrue(node.progress.is_started)
      self.assertTrue(node.progress.is_completed)
      if node.is_leaf:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_completed, 10)
        self.assertEqual(node.progress.num_failed, 1)
      else:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_failed, 0)
        self.assertEqual(node.progress.num_processed, node.progress.num_total)

  def test_concurrent_startup_delay(self):
    plugin = TestPlugin()
    exp = eval_test_helper.test_experiment()
    root_dir = os.path.join(
        tempfile.mkdtemp(), 'test_concurrent_startup_delay'
    )
    _ = exp.run(
        root_dir,
        runner='parallel',
        plugins=[plugin],
        concurrent_startup_delay=(0, 5),
    )


class DebugRunnerTest(RunnerTest):

  def test_debug_runner(self):
    plugin = TestPlugin()
    exp = eval_test_helper.test_experiment()
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_debug_runner')
    run = exp.run(root_dir, runner='debug', plugins=[plugin])

    self.assertIsNotNone(plugin.start_time)
    self.assertIsNotNone(plugin.complete_time)
    self.assertGreater(plugin.complete_time, plugin.start_time)

    self.assertEqual(
        len(plugin.started_experiments), len(exp.nodes)
    )
    self.assertEqual(
        len(plugin.completed_experiments), len(exp.nodes)
    )
    self.assertEqual(
        len(plugin.started_example_ids), 6 * 1
    )
    self.assertEqual(
        len(plugin.completed_example_ids), 6 * 1
    )
    self.assert_same_list(plugin.skipped_experiments, [])
    self.assert_same_list(plugin.skipped_example_ids, [])
    self.assertFalse(
        pg.io.path_exists(os.path.join(run.output_root, 'run.json'))
    )

    for node in exp.nodes:
      self.assertTrue(node.progress.is_started)
      self.assertTrue(node.progress.is_completed)
      if node.is_leaf:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_completed, 1)
        self.assertEqual(node.progress.num_failed, 0)
      else:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_failed, 0)
        self.assertEqual(node.progress.num_processed, node.progress.num_total)


if __name__ == '__main__':
  unittest.main()
